//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtkmm.h>
#include <glibmm.h>
#include <glibmm/i18n.h>
#include <glib/gstdio.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <unistd.h>
#include <sys/stat.h>

#include "util.hpp"
#include "uri++.hpp"
#include "file_utils.hpp"
#include "vfs.hpp"
#include "network.hpp"

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlreader.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include <boost/format.hpp>

#include "xml.hpp"
#include "main.hpp"
#include "paths.hpp"

#include "x_library.hpp"
#include "x_vfs.hpp"

#define AMAZON_FORMAT_STRING_URI "http://webservices.amazon.com/onca/xml?Service=AWSECommerceService&SubscriptionId=1E90RVC80K4MVNTCHG02&Operation=ItemSearch&Artist=%s&Title=%s&SearchIndex=Music&ResponseGroup=Images&Version=2004-11-10"
#define AMAZON_FORMAT_STRING_URI_ASIN "http://images.amazon.com/images/P/%s.01.MZZZZZZZ.jpg"

namespace Bmp
{
  namespace Amazon
  {
    enum Exception
    {
      NO_COVER
    };

    void 
    get_cover_by_asin (const std::string& asin, Glib::RefPtr<Gdk::Pixbuf>& cover)
    {
	if (!Bmp::Network::is_connected ()) throw NO_COVER;

	Bmp::URI uri ((boost::format (AMAZON_FORMAT_STRING_URI_ASIN) % asin).str());
	uri.escape ();

	Bmp::VFS::Handle handle = Bmp::VFS::Handle::create (Glib::ustring(uri));
	vfs->read_no_container (handle);
    
	if (!handle.get_buffer())
	{
	  throw NO_COVER;
	}

	Glib::RefPtr<Gdk::PixbufLoader> loader = Gdk::PixbufLoader::create ("jpeg"); 
	loader->write (reinterpret_cast<const guint8*>(handle.get_buffer()), handle.get_buffer_size());
	loader->close ();
	cover = loader->get_pixbuf()->scale_simple ( 128, 128, Gdk::INTERP_HYPER );
    }

    void 
    get_cover_artist_album (const std::string& artist_, const std::string& album_, Glib::RefPtr<Gdk::Pixbuf>& cover, bool only_cached) 
    {
	if (!Bmp::Network::is_connected ()) throw NO_COVER;

	xmlDocPtr doc;
	xmlXPathObjectPtr xpathObj;
	xmlNodeSetPtr nv;
	xmlNodePtr node;
	char *image_uri = 0;

	Glib::ustring artist = artist_; 
	Glib::ustring album  = album_; 

	std::string basename = (boost::format ("%1%_%2%.png") % artist.lowercase () % album.lowercase ()).str ();
	std::string thumb_path = Glib::build_filename (BMP_PATH_COVER_THUMB_DIR, basename);

	if (Glib::file_test (thumb_path, Glib::FILE_TEST_EXISTS))
	{
	      cover = Gdk::Pixbuf::create_from_file (thumb_path); 
	      return;
	}
	else if (!only_cached)
	{
	      Bmp::URI uri ((boost::format (AMAZON_FORMAT_STRING_URI) % artist % album).str());
	      uri.escape ();

	      Bmp::VFS::Handle handle = Bmp::VFS::Handle::create (Glib::ustring(uri));
	      vfs->read_no_container (handle);

	      if (!handle.get_buffer())
	      {
		throw NO_COVER;
	      }

	      doc = xmlParseMemory ((const char*)handle.get_buffer(), handle.get_buffer_size());
	      if (!doc)
	      {
		throw NO_COVER;
	      }

	      xpathObj = xml_execute_xpath_expression (doc, BAD_CAST "//amazon:MediumImage/amazon:URL", 
							    BAD_CAST "amazon=http://webservices.amazon.com/AWSECommerceService/2004-11-10");

	      nv = xpathObj->nodesetval;

	      if ((!nv) || (!nv->nodeNr) || (!nv->nodeTab[0]->children))
	      {
		xmlXPathFreeObject (xpathObj);
		throw NO_COVER;
	      }

	      node = nv->nodeTab[0];
	      image_uri = (char*)XML_GET_CONTENT (node->children);
	      xmlXPathFreeObject (xpathObj);

	      handle = Bmp::VFS::Handle::create (Glib::ustring(image_uri));
	      free (image_uri);
	      vfs->read_no_container (handle);

	      if (!handle.get_buffer())
	      {
		throw NO_COVER; 
	      }

	      Glib::RefPtr<Gdk::PixbufLoader> loader = Gdk::PixbufLoader::create ("jpeg"); 
	      loader->write (reinterpret_cast<const guint8*>(handle.get_buffer()), handle.get_buffer_size());
	      loader->close ();
	      cover = loader->get_pixbuf()->scale_simple ( 128, 128, Gdk::INTERP_HYPER );
	      cover->save (thumb_path, "png");
	      return;
	}

	throw NO_COVER;
    }

    void 
    get_cover (const std::string& _uri, Glib::RefPtr<Gdk::Pixbuf>& cover, bool only_cached) 
    {
	if (!Bmp::Network::is_connected ()) throw NO_COVER;

	xmlDocPtr doc;
	xmlXPathObjectPtr xpathObj;
	xmlNodeSetPtr nv;
	xmlNodePtr node;
	char *image_uri = 0;

	Bmp::DB::DataRow row;
	library->get_metadata (_uri, row); 

	Bmp::Library::Track track (row);

	if (!track.artist)
	    throw NO_COVER;
	Glib::ustring artist = track.artist.get();

	if (!track.album)
	    throw NO_COVER;
	Glib::ustring album  = track.album.get();

	if (!track.location)
	    throw NO_COVER;

	Glib::ustring location = track.location.get(); 
	Bmp::URI uri (track.location.get());

	if ((album.length () > 0) && (uri.get_protocol() != Bmp::URI::PROTOCOL_HTTP)) 
	{
	    std::string basename = (boost::format ("%1%_%2%.png") % artist.lowercase () % album.lowercase ()).str ();
	    std::string thumb_path = Glib::build_filename (BMP_PATH_COVER_THUMB_DIR, basename);

	    if (Glib::file_test (thumb_path, Glib::FILE_TEST_EXISTS))
	    {
		cover = Gdk::Pixbuf::create_from_file (thumb_path); 
		return;
	    }
	    else if (!only_cached)
	    {
	      Bmp::URI uri ((boost::format (AMAZON_FORMAT_STRING_URI) % artist % album).str());
	      uri.escape ();

	      Bmp::VFS::Handle handle = Bmp::VFS::Handle::create (Glib::ustring(uri));
	      vfs->read_no_container (handle);

	      doc = xmlParseMemory ((const char*)handle.get_buffer(), handle.get_buffer_size());
	      if (!doc)
	      {
		throw NO_COVER;
	      }

	      xpathObj = xml_execute_xpath_expression (doc, BAD_CAST "//amazon:MediumImage/amazon:URL", 
							    BAD_CAST "amazon=http://webservices.amazon.com/AWSECommerceService/2004-11-10");

	      nv = xpathObj->nodesetval;

	      if ((!nv) || (!nv->nodeNr) || (!nv->nodeTab[0]->children))
	      {
		xmlXPathFreeObject (xpathObj);
		throw NO_COVER;
	      }

	      node = nv->nodeTab[0];
	      image_uri = (char*)XML_GET_CONTENT (node->children);
	      xmlXPathFreeObject (xpathObj);

	      handle = Bmp::VFS::Handle::create (Glib::ustring(image_uri));
	      free (image_uri);
	      vfs->read_no_container (handle);

	      if (!handle.get_buffer())
	      {
		throw NO_COVER; 
	      }

	      Glib::RefPtr<Gdk::PixbufLoader> loader = Gdk::PixbufLoader::create ("jpeg"); 
	      loader->write (reinterpret_cast<const guint8*>(handle.get_buffer()), handle.get_buffer_size());
	      loader->close ();

	      cover = loader->get_pixbuf()->scale_simple ( 128, 128, Gdk::INTERP_HYPER );
	      cover->save (thumb_path, "png");
	      return;
	    }
	}
	throw NO_COVER;
    }
  }
}
