#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <boost/format.hpp>

#include "file_utils.hpp"
#include "ui_util.hpp"
#include "util.hpp"
#include "paths.hpp"
#include "main.hpp"

#include "library_ui.hpp"
#include "library_ui_merge.hpp"
#include "musicbrainz_ws.hpp"

#include "x_library.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

namespace Bmp
{
      LibraryUI::LibraryUI (BaseObjectType                        *cobject,
			    const Glib::RefPtr<Gnome::Glade::Xml> &xml)
	: Gtk::Window (cobject),
	  ref_xml (xml),
	  updated (false)
      {
	Gtk::Image *image;

	bmp_window_set_icon_list (GTK_WIDGET (gobj ()), "library");

	ref_xml->get_widget ("i_header", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "library" G_DIR_SEPARATOR_S "header-logo.png"));

	ref_xml->get_widget ("i_add", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "library" G_DIR_SEPARATOR_S "add.png"));
	ref_xml->get_widget ("i_wait1", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "wait.gif"));


	ref_xml->get_widget ("i_remove_artist", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "library" G_DIR_SEPARATOR_S "remove.png"));
	ref_xml->get_widget ("i_remove_album", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "library" G_DIR_SEPARATOR_S "remove.png"));
	ref_xml->get_widget ("i_remove_tracks", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "library" G_DIR_SEPARATOR_S "remove.png"));


	ref_xml->get_widget ("b_add",
			      b_add);
	b_add->signal_clicked().connect			(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::music_add));

	ref_xml->get_widget ("b_remove_tracks",
			      b_remove_tracks);
	b_remove_tracks->signal_clicked().connect       (sigc::mem_fun (*this,
							      &Bmp::LibraryUI::remove_tracks));

	ref_xml->get_widget ("b_remove_album",
			      b_remove_album);
	b_remove_album->signal_clicked().connect       (sigc::mem_fun (*this,
							      &Bmp::LibraryUI::remove_album));

	ref_xml->get_widget ("b_remove_artist",
			      b_remove_artist);
	b_remove_artist->signal_clicked().connect       (sigc::mem_fun (*this,
							      &Bmp::LibraryUI::remove_artist));

	ref_xml->get_widget ("b_close",
			      b_close);
	b_close->signal_clicked().connect		(sigc::mem_fun (*this,
							      &Gtk::Widget::hide));

	ref_xml->get_widget ("b_play",
			      b_play); 
	b_play->signal_clicked().connect		(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::play));

	ref_xml->get_widget ("b_enqueue",
			      b_enqueue);
	b_enqueue->signal_clicked().connect		(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::enqueue));
	ref_xml->get_widget ("b_select_all",
			      b_select_all);
	b_select_all->signal_clicked().connect		(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::select_all));

#ifdef USE_AMAZON
	ref_xml->get_widget ("b_fetch_cover",
			      b_fetch_cover);
	b_fetch_cover->signal_clicked().connect		(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::fetch_cover));
#endif

	ref_xml->get_widget ("b_merge_albums",
			      b_merge_albums);
	b_merge_albums->signal_clicked().connect	(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::merge_albums));

	// Search widgets
	using namespace Bmp::Library;
	using namespace Gtk;

	const Datum search_data[] = {
	    DATUM_ARTIST,
	    DATUM_ALBUM,
	    DATUM_TITLE,
	    DATUM_GENRE,
	};
  
	cs_store = Gtk::ListStore::create (cr_search);

	for (unsigned int n = 0; n < G_N_ELEMENTS (search_data); ++n) 
	{
	  TreeModel::iterator i = cs_store->append ();
	  DatumDefine define = library->get_metadatum_define (search_data[n]);
	  (*i)[cr_search.name] = define.title; 
	  (*i)[cr_search.id] = define.id;
	}

	static boost::format f_cbox     ("cbox_%d");
	static boost::format f_cbutton  ("cb_%d");
	static boost::format f_entry    ("entry_%d");
	for (int n = 0; n < N_FILTERS; ++n)
	{
	    ref_xml->get_widget ((f_cbutton % (n+1)).str(), cbutton[n]);
	    ref_xml->get_widget ((f_cbox    % (n+1)).str(), cbox[n]);
	    ref_xml->get_widget ((f_entry   % (n+1)).str(), entry[n]);

	    cbutton[n]->signal_toggled ().connect (sigc::bind (sigc::mem_fun (*this, &Bmp::LibraryUI::filter_toggled), FilterID(n)));
	    cbox[n]->clear ();
	    Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
	    cbox[n]->pack_start (*cell);
	    cbox[n]->add_attribute (*cell, "text", 0);
	    cbox[n]->set_model (cs_store);
	}

	// Watch shutdown-request from SystemControl
	g_signal_connect (G_OBJECT (bmp_system_control), "shutdown-request", G_CALLBACK (Bmp::LibraryUI::shutdown_request), this);
	shutdown.lock ();

	ref_xml->get_widget ("notebook_add",  notebook_add);
	ref_xml->get_widget ("view_tracks",   view_tracks);
	ref_xml->get_widget ("view_albums",   view_albums);
	ref_xml->get_widget ("albums_filter", albums_filter);

	//Tracks view
	struct { 
	    char    *title;
	    double   xalign;
	    int	     width;
	} headers_tracks[] = {
	    { _("Track"),      1.0, 50  },
	    { _("Title"),      0.0, 150 },
	    { _("Genre"),      0.0, 100 },
	    { _("Bitrate"),    0.0, 80  },
	    { _("Samplerate"), 0.0, 80  },
	};

	for (unsigned int n = 0; n < G_N_ELEMENTS(headers_tracks); ++n)
	{
	  Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );

	  cell->property_ellipsize () = Pango::ELLIPSIZE_END;
	  cell->property_xalign() = headers_tracks[n].xalign;

	  view_tracks->append_column (headers_tracks[n].title, *cell);
	  Gtk::TreeViewColumn *column = view_tracks->get_column (n);

	  column->add_attribute (*cell, "text", n);
	  column->set_resizable (true);
	  column->set_sort_column (n);
	  column->set_expand (true);
	  column->property_min_width() = headers_tracks[n].width;
	}

	store_tracks = Gtk::ListStore::create (tracks);

	view_tracks->set_model (store_tracks);
	view_tracks->get_selection()->set_mode (Gtk::SELECTION_MULTIPLE);
	view_tracks->get_selection()->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::tracks_selection_changed));


	//Albums view
	enum Renderer
	{
	  R_TEXT,
	  R_PIXBUF,
	};

	struct { 
	    char    *title;
	    double   xalign;
	    int	     width;
	    Renderer renderer;
	} headers_albums[] = {
	    { _("Cover"),      0.5, 72,  R_PIXBUF },
	    { _("Album"),      0.0, 250, R_TEXT   },
	};

	for (unsigned int n = 0; n < G_N_ELEMENTS(headers_albums); ++n)
	{
	  switch (headers_albums[n].renderer)
	  {
		case R_PIXBUF:
		{
		  Gtk::CellRendererPixbuf *cell = Gtk::manage ( new Gtk::CellRendererPixbuf() );

		  cell->property_xalign() = headers_albums[n].xalign;
		  cell->property_xpad() = 3;
		  cell->property_ypad() = 4;

	  	  view_albums->append_column (headers_albums[n].title, *cell);
		  Gtk::TreeViewColumn *column = view_albums->get_column (n);

	  	  column->add_attribute (*cell, "pixbuf", n);
	  	  column->set_resizable (false);
	  	  column->set_expand (false);
	  	  column->set_sizing (Gtk::TREE_VIEW_COLUMN_FIXED);
		  column->property_min_width() = headers_albums[n].width;
		  column->property_max_width() = headers_albums[n].width;

		  break;
		}

		case R_TEXT:
		{
		  Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );

		  cell->property_ellipsize () = Pango::ELLIPSIZE_END;
		  cell->property_xalign() = headers_albums[n].xalign;
		  cell->property_size() = 14 * PANGO_SCALE;
		  cell->property_width () = 700;

	  	  view_albums->append_column (headers_albums[n].title, *cell);
		  Gtk::TreeViewColumn *column  = view_albums->get_column (n);

	  	  column->add_attribute (*cell, "markup", n);
	  	  column->set_resizable (false);
	  	  column->set_expand (false);
	  	  column->set_sizing (Gtk::TREE_VIEW_COLUMN_FIXED);
		  column->property_min_width() = headers_albums[n].width;
		  column->property_max_width() = headers_albums[n].width;

		  break;
		}
	  }

	}

	store_albums = Gtk::ListStore::create (albums);
	store_albums->set_default_sort_func (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_default_sort_func));
	store_albums->set_sort_column_id (-1, Gtk::SORT_ASCENDING);

	store_albums_filter = Gtk::TreeModelFilter::create (store_albums);
	store_albums_filter->set_visible_func (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_row_is_visible));

	view_albums->set_model (store_albums_filter);
	view_albums->set_fixed_height_mode ();
	view_albums->get_selection()->set_mode (Gtk::SELECTION_MULTIPLE);
	view_albums->get_selection()->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_selection_changed));
	view_albums->signal_button_press_event ().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_button_press));
	view_albums->signal_row_activated ().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::play_current));

	albums_filter->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_refilter));

	//Connect to the library for blocking the UI to avoid race conditions
	library->signal_processing_start().connect (sigc::bind (sigc::mem_fun (*this, &Gtk::Widget::set_sensitive), false));
	library->signal_processing_end().connect (sigc::bind (sigc::mem_fun (*this, &Gtk::Widget::set_sensitive), true));
      }

      LibraryUI*
      LibraryUI::create ()
      {
	const std::string path = Glib::build_filename (DATA_DIR, "glade" G_DIR_SEPARATOR_S "library_ui.glade");
	Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
	LibraryUI *library_ui = 0;
	glade_xml->get_widget_derived ("library_ui", library_ui);
	return library_ui;
      }

      //////////////////////////// Search 
      void
      LibraryUI::filter_toggled (FilterID id)
      {
	bool active = cbutton[int(id)]->get_active();
	cbox[int(id)]->set_sensitive (active);
	entry[int(id)]->set_sensitive (active);
      }

      //////////////////////////// Handling of covers/cover-fetching
#ifdef USE_AMAZON
      //FIXME: If not built with amazon support, the entire fetch-cover widget should be probably hidden
      void
      LibraryUI::fetch_cover ()
      { 
	using namespace Gtk;
	using namespace Gdk;
	using namespace Glib;

	TreeSelection::ListHandle_Path paths = view_albums->get_selection()->get_selected_rows ();
	TreeModel::iterator i_Selected = store_albums_filter->convert_iter_to_child_iter (store_albums_filter->get_iter (*(paths.begin()))); 

	ustring key_A = (*i_Selected)[albums.artist];
	ustring key_B = (*i_Selected)[albums.album];

	this->set_sensitive (false); //FIXME: Don't set the entire window insensitive

	RefPtr<Pixbuf> cover;
	try {
	    Amazon::get_cover_artist_album (key_A, key_B, cover, false);
	    (*i_Selected)[albums.cover] = cover->scale_simple (64, 64, INTERP_BILINEAR);
	    (*i_Selected)[albums.has_cover] = true;
	    b_fetch_cover->set_sensitive (false);
	} catch (...) {}

	this->set_sensitive (true);
      }
#endif

      bool
      LibraryUI::albums_button_press (GdkEventButton* event)
      {
	Gtk::TreeViewColumn *column, *column_0 = view_albums->get_column (0);
	Gtk::TreeModel::Path path;
	int cell_x, cell_y;

	g_message ("x: %d, y: %d", int(event->x), int(event->y));

	if (view_albums->get_path_at_pos (int(event->x),
					  int(event->y),
					  path,
					  column,
					  cell_x,
					  cell_y)) 
	{
	  g_message ("column: %p, column_0: %p", gpointer(column), gpointer(column_0));
	  if (column == column_0) return true; 
	}
      
	return false;
      } 

      //////////////////////////// Sorting
      int 
      LibraryUI::albums_default_sort_func (const Gtk::TreeModel::iterator& iter_a, const Gtk::TreeModel::iterator& iter_b)
      {
	std::string str_a = ((*iter_a)[albums.artist] + " " + (*iter_a)[albums.album]);
	std::string str_b = ((*iter_b)[albums.artist] + " " + (*iter_b)[albums.album]);

	return str_a.compare(str_b);
      }

      //////////////////////////// Selections
      void
      LibraryUI::tracks_selection_changed ()
      {
	  bool selected = view_tracks->get_selection ()->count_selected_rows ();

	  b_play-> set_sensitive (selected);
	  b_enqueue-> set_sensitive (selected);
	  b_remove_tracks-> set_sensitive (selected);
      }

      void
      LibraryUI::albums_selection_changed ()
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Gtk;

          store_tracks->clear ();
	  b_remove_album->set_sensitive (false);

	  if (!view_albums->get_selection ()->count_selected_rows ())
	  {
#ifdef USE_AMAZON
	    b_fetch_cover->set_sensitive (false);
#endif
	    return;
	  }

          TreeSelection::ListHandle_Path paths = view_albums->get_selection()->get_selected_rows ();

	  if (paths.size () == 1)
	  {
	    //FIXME: We should possibly allow for removal of several albums in a run
	    b_remove_album->set_sensitive (true);
	    b_merge_albums->set_sensitive (false);
#ifdef USE_AMAZON
	    //FIXME: This is a little awkward as we do an iteration below anyway
	    //FIXME: We should possibly allow for fetching of several albums in a run
	    TreeModel::iterator cur = store_albums_filter->get_iter (*(paths.begin())); 
	    b_fetch_cover->set_sensitive (!(*cur)[albums.has_cover]);
#endif
	  }
	  else
	  {
	    b_merge_albums->set_sensitive (true);
	  }

	  std::string id_artist = library->get_metadatum_id (DATUM_ARTIST);
	  std::string id_album  = library->get_metadatum_id (DATUM_ALBUM);

          for (TreeSelection::ListHandle_Path::const_iterator path = paths.begin (); path != paths.end (); ++path)
          {
	    TreeModel::iterator cur = store_albums_filter->get_iter (*path);

	    Glib::ustring key_A = (*cur)[albums.artist];
	    Glib::ustring key_B = (*cur)[albums.album];

	    AttributeList attrs; 
	    DataRowV vector;

	    attrs.push_back (Attribute (true, id_artist, std::string(key_A)));
	    attrs.push_back (Attribute (true, id_album,  std::string(key_B)));

	    library->query (attrs, vector);

            for (DataRowV::iterator v = vector.begin (), end = vector.end (); v != end; ++v) 
            {
              TreeModel::iterator tracks_iter = store_tracks->append ();
	      Track track (*v);

              if (track.title) (*tracks_iter)[tracks.title] = track.title.get (); 
              if (track.tracknumber) (*tracks_iter)[tracks.track] = track.tracknumber.get (); 
              if (track.genre) (*tracks_iter)[tracks.genre] = track.genre.get (); 
              if (track.bitrate) (*tracks_iter)[tracks.bitrate] = track.bitrate.get (); 
              if (track.samplerate) (*tracks_iter)[tracks.samplerate] = track.samplerate.get (); 
              if (track.location) (*tracks_iter)[tracks.uri] = track.location.get (); 
            }

	  }

	  if (!store_tracks->children ().empty ())
	  {
	    b_select_all->set_sensitive (true);
	  }

      }

      void
      LibraryUI::select_all ()
      {
	view_tracks->get_selection()->select_all ();
      }

      //////////////////////////// Removing
      void
      LibraryUI::remove_artist ()
      {
      }

      void
      LibraryUI::remove_album ()
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Glib;
	  using namespace Gtk;
	  using namespace std;

	  TreeSelection::ListHandle_Path paths = view_albums->get_selection()->get_selected_rows ();
	  TreeModel::iterator selected = store_albums_filter->convert_iter_to_child_iter (store_albums_filter->get_iter (*(paths.begin()))); 

          AttributeList attrs;

          ustring key_A = (*selected)[albums.artist];
          ustring key_B = (*selected)[albums.album];

          attrs.push_back (Attribute (true, library->get_metadatum_id (DATUM_ARTIST), string(key_A)));
          attrs.push_back (Attribute (true, library->get_metadatum_id (DATUM_ALBUM),  string(key_B)));

          library->del (attrs);
          store_albums->erase (selected);
      }

      void
      LibraryUI::remove_tracks ()
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Glib;
	  using namespace Gtk;
	  using namespace std;

	  typedef std::vector<Gtk::TreeModel::iterator> IterList;
	  IterList iters;

          TreeSelection::ListHandle_Path paths = view_tracks->get_selection()->get_selected_rows ();
          for (TreeSelection::ListHandle_Path::const_iterator iter = paths.begin (); iter != paths.end (); ++iter)
	  {
	      iters.push_back (store_tracks->get_iter (*iter));
	  }

          for (IterList::const_iterator cur = iters.begin (), end = iters.end (); cur != end; ++cur)
          {
              ustring key = (*(*cur))[tracks.uri];
              AttributeList attrs;
              attrs.push_back (Attribute (true, library->get_metadatum_id (DATUM_LOCATION), string(key)));
              library->del (attrs);
	      store_tracks->erase (*cur);
          }

          if (store_tracks->children ().empty())
          {
              remove_album ();
          }
      }

      //////////////////////////// Play/Enqueue
      void
      LibraryUI::play_current (const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
      {
	  select_all ();
	  play ();
      }

      void
      LibraryUI::play ()
      {
	  Gtk::TreeSelection::ListHandle_Path paths = view_tracks->get_selection()->get_selected_rows ();
	  char **uri_list = (char**)calloc (paths.size()+1, sizeof(char*));
	  unsigned int counter = 0;
	  for (Gtk::TreeSelection::ListHandle_Path::const_iterator iter = paths.begin (); iter != paths.end (); ++iter)
	  {
	      Gtk::TreeModel::iterator cur = store_tracks->get_iter (*iter);
	      Glib::ustring str = (*cur)[tracks.uri];
	      uri_list[counter++] = strdup(str.c_str ());
	  }
	  uri_list[paths.size()] = 0;

	  int out;
	  bmp_system_control_add_uri_list (bmp_system_control,
					   (const char**)uri_list,
					   -1,
					   true, 
					   true, 
					   0, 
					   &out,
					   NULL);
	  g_strfreev (uri_list);
	  view_tracks->get_selection()->unselect_all ();
      }

      void
      LibraryUI::enqueue ()
      {
	  Gtk::TreeSelection::ListHandle_Path paths = view_tracks->get_selection()->get_selected_rows ();
	  char **uri_list = (char**)calloc (paths.size()+1, sizeof(char*));
	  unsigned int counter = 0;
	  for (Gtk::TreeSelection::ListHandle_Path::const_iterator iter = paths.begin (); iter != paths.end (); ++iter)
	  {
	      Gtk::TreeModel::iterator cur = store_tracks->get_iter (*iter);
	      Glib::ustring str = (*cur)[tracks.uri];
	      uri_list[counter++] = strdup(str.c_str ());
	  }
	  uri_list[paths.size()] = 0;

	  int out;
	  bmp_system_control_add_uri_list (bmp_system_control,
					   (const char**)uri_list,
					   -1,
					   false, 
					   false, 
					   -1, 
					   &out,
					   NULL);
	  g_strfreev (uri_list);
	  view_tracks->get_selection()->unselect_all ();
      }

      //////////////////////////// Adding Music/Rebuilding view 
      void
      LibraryUI::merge_albums ()
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Gtk;
	  using namespace Glib;

	  if (view_albums->get_selection ()->count_selected_rows () < 2) //FIXME: Doesn't make sense to "group" only 1 album or none
	  {
	    return;
	  }

	  Util::FileList list, non_writable;
	  
	  TreeModel::Children nodes = store_tracks->children ();
          for (TreeModel::Children::const_iterator node = nodes.begin (), end = nodes.end (); node != end; ++node)
          {
	    Glib::ustring uri = (*node)[tracks.uri];
	    list.push_back (uri);
	  }
	  Util::files_writable (list, non_writable);
	  if (!non_writable.empty ())
	  {
	    Gtk::MessageDialog dialog (*this, "Unable to merge: Some files are not writable.", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
	    dialog.run ();
	    return;
#if 0
	    for (Util::FileList::iterator file = non_writable.begin (), end = non_writable.end (); file != end; ++file)
	    {
		g_message ("File without write permission: '%s'", (*file).c_str()); 
	    }
#endif
	  }

	  typedef std::vector < std::string > Albums; 
	  typedef TreeModel::iterator Iter;
	  typedef TreeSelection::ListHandle_Path Paths;
      
	  MusicBrainzWS::MB_Releases mb_releases; 
	  Paths paths = view_albums->get_selection()->get_selected_rows ();
	  for (Paths::const_iterator path  = paths.begin (),
				     end   = paths.end ();
				     path != end; ++path)
	  {
	    Iter cur = store_albums_filter->convert_iter_to_child_iter (store_albums_filter->get_iter (*path)); 
	    ustring artist = (*cur)[albums.artist];
	    ustring album  = (*cur)[albums.album];

	    MusicBrainzWS::get_releases (artist, album, mb_releases);
	  }

	  LibraryUIMergeDialog *dialog = LibraryUIMergeDialog::create ();
	  dialog->run (mb_releases);
	  delete dialog;
      }

      void
      LibraryUI::music_add ()
      {
	Gtk::FileChooserDialog dialog (*this,
                                       _("Select Path To Add"),
                                       Gtk::FILE_CHOOSER_ACTION_SELECT_FOLDER);

	dialog.add_button (Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	dialog.add_button (Gtk::Stock::ADD, Gtk::RESPONSE_OK);
	dialog.set_current_folder (mcs->key_get<std::string>("bmp", "file-chooser-path"));

	if (dialog.run () == Gtk::RESPONSE_CANCEL)
	{
	   return;
	}
	else
	{
	  notebook_add->set_current_page (1);
	  Glib::ustring folder = dialog.get_filename ();
	  dialog.hide ();

	  Util::FileList list;
	  Util::collect_path (folder, list);

	  for (Util::FileList::const_iterator iter = list.begin (), end = list.end ();
	       iter != end;
	       ++iter)
	  {
	    library->cache_metadata (*iter, folder);
	    while (gtk_events_pending ()) gtk_main_iteration ();
	  }

	  library->update_compilations ();

	  notebook_add->set_current_page (0);
	  update_view ();
	}
    }

    void
    LibraryUI::albums_refilter ()
    {
	store_albums_filter->refilter ();
    }

    bool
    LibraryUI::albums_row_is_visible (const Gtk::TreeModel::const_iterator& iter)
    {
	Glib::ustring artist ((*iter)[albums.artist]);	
	Glib::ustring album  ((*iter)[albums.album]);	
	std::string needle   (albums_filter->get_text ());
  
	return (needle.empty() || match_keys_boost (artist + " " + album, needle));
    }

    void
    LibraryUI::update_view ()
    {
	using namespace Glib; 
	using namespace Gtk; 
	using namespace Gdk; 
	using namespace Bmp::DB;
	using namespace Bmp::Library;
	using namespace std;
	using namespace boost;

	bool break_loop = false;

	Glib::RefPtr<Gdk::Pixbuf> disc = 
		Gdk::Pixbuf::create_from_file (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "audio.png"))->scale_simple 
					  (64, 64, Gdk::INTERP_BILINEAR);

        string id_album    = library->get_metadatum_id (DATUM_ALBUM);
        string id_artist   = library->get_metadatum_id (DATUM_ARTIST);
	string id_location = library->get_metadatum_id (DATUM_LOCATION);

	store_albums->clear ();
    

        AttributeList attrs_albums;
        attrs_albums.push_back (Attribute (false, id_location, std::string("file://"))); //Make sure we only fetch files

        DataRowV vector_albums;
        library->project (id_album, vector_albums, attrs_albums);

        for (DataRowV::iterator iter = vector_albums.begin (), end = vector_albums.end (); iter != end; ++iter)
	{
	  if (break_loop) break;

          string album = get<std::string>((*iter)[id_album]);
          if (album.length())
          {
	      if (library->is_compilation (album)) continue;

	      AttributeList attrs_artists;
	      attrs_artists.push_back (Attribute (false, id_location, string("file://")));
	      attrs_artists.push_back (Attribute (true,  id_album, album));

	      DataRowV vector_artists;
	      library->project (id_artist, vector_artists, attrs_artists);

	      for (DataRowV::iterator iter = vector_artists.begin (), end = vector_artists.end (); iter != end; ++iter)
	      {
		string artist = get<std::string>((*iter)[id_artist]);

		if (artist.length())
		{
		  TreeModel::iterator cur = store_albums->append ();
#ifdef USE_AMAZON
		  RefPtr<Pixbuf> cover;
		  try {
		    Amazon::get_cover_artist_album (artist, album, cover, true);
		    (*cur)[albums.cover] = cover->scale_simple (64, 64, INTERP_BILINEAR);
		    (*cur)[albums.has_cover] = true;
		  }
		  catch (...)
		  {
		    (*cur)[albums.cover] = disc;
		    (*cur)[albums.has_cover] = false;
		  }

#else
		  (*cur)[albums.cover]  = disc;
#endif

		  (*cur)[albums.album]  = album;
		  (*cur)[albums.artist] = artist;

		  stringstream title;
		  title << Markup::escape_text (artist) << "\n" << "<i>" << Markup::escape_text (album) << "</i>";
		  (*cur)[albums.title]  = title.str(); 

		  while (gtk_events_pending ()) gtk_main_iteration ();

		  if (shutdown.trylock())
		  {
		    shutdown.unlock ();
		    break_loop = true;
		    break;
		  }


		}
	      }
          }
	}
    }

}
