#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <boost/format.hpp>

#include "file_utils.hpp"
#include "util.hpp"
#include "paths.hpp"
#include "main.hpp"

#include "library_ui_merge.hpp"
#include "musicbrainz_ws.hpp"

#include "x_library.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

namespace Bmp
{
      LibraryUIMergeDialog::LibraryUIMergeDialog (BaseObjectType *cobject,
			    const Glib::RefPtr<Gnome::Glade::Xml> &xml)
	: Gtk::Dialog (cobject),
	  ref_xml (xml),
	  coverart_default (Gdk::Pixbuf::create_from_file (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "audio.png")))
      {
	  ref_xml->get_widget ("merge_cb_artist",   merge_cb_artist);
	  ref_xml->get_widget ("merge_cb_album",    merge_cb_album);
	  ref_xml->get_widget ("merge_coverart",    merge_coverart);
	  ref_xml->get_widget ("merge_album",	    merge_album);
	  ref_xml->get_widget ("merge_artist",	    merge_artist);
	  ref_xml->get_widget ("merge_artist_id",   merge_artist_id);
	  ref_xml->get_widget ("merge_ASIN",	    merge_ASIN);
	  ref_xml->get_widget ("merge_release_ID",  merge_release_ID);
	  ref_xml->get_widget ("merge_score",	    merge_score);
      }

      LibraryUIMergeDialog*
      LibraryUIMergeDialog::create ()
      {
	const std::string path = Glib::build_filename (DATA_DIR, "glade" G_DIR_SEPARATOR_S "library_ui_merge.glade");
	Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
	LibraryUIMergeDialog *library_ui = 0;
	glade_xml->get_widget_derived ("dialog_merge_albums", library_ui);
	return library_ui;
      }

      int
      LibraryUIMergeDialog::run (const MusicBrainzWS::MB_Releases& releases_)
      {
	  using namespace Bmp::MusicBrainzWS;
	  using namespace Gtk;

	  merge_cb_artist_store = ListStore::create (artists);
	  merge_cb_album_store  = ListStore::create (albums);

	  for (MB_Releases::const_iterator r_Iter     = releases_.begin (),
					   r_IterEnd  = releases_.end ();
					   r_Iter    != r_IterEnd; ++r_Iter)
	  {

	    MB_Release const& release = (*r_Iter);
	    merge_releases.insert (std::make_pair (release.id, release));
  
	    ArtistReleaseMap::iterator rm_Iter = merge_artist_release_map.find (release.artist_id);
	    if (rm_Iter != merge_artist_release_map.end ())
	    {
	      ReleaseIds &ids = rm_Iter->second;
	      ids.push_back (release.id);
	    }
	    else
	    {
	      TreeModel::iterator m_Iter  = merge_cb_artist_store->append ();

	      (*m_Iter)[artists.name]	  = release.artist_name;
	      (*m_Iter)[artists.id]	  = release.artist_id;
	      
	      ReleaseIds ids; 
	      ids.push_back (release.id);
	      merge_artist_release_map.insert (std::make_pair (release.artist_id, ids)); 
	    }
	  }

	  CellRendererText *cell;

	  // Setup 'Artist' ComboBox 
	  cell = Gtk::manage (new CellRendererText ());
	  merge_cb_artist->clear ();
	  merge_cb_artist->pack_start (*cell);
	  merge_cb_artist->add_attribute (*cell, "text", 0);
	  merge_cb_artist->set_model (merge_cb_artist_store);
	  merge_cb_artist->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIMergeDialog::artist_changed));

	  // Setup 'Album' ComboBox 
	  cell = Gtk::manage (new CellRendererText ());
	  merge_cb_album->clear ();
	  merge_cb_album->pack_start (*cell);
	  merge_cb_album->add_attribute (*cell, "text", 0);
	  merge_cb_album->set_model (merge_cb_album_store);
	  merge_cb_album->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIMergeDialog::album_changed));

	  return Gtk::Dialog::run ();
      }

      void
      LibraryUIMergeDialog::album_changed ()
      {
	  using namespace Gtk;
	  using namespace MusicBrainzWS;

	  this->set_sensitive (false);
	
	  // Get releases from selected artist
	  TreeModel::const_iterator m_Iter = merge_cb_album->get_active ();
	  ReleaseMap::iterator r_Iter = merge_releases.find ((*m_Iter)[albums.id]);
	  MB_Release const& release = r_Iter->second;

	  merge_coverart->set (coverart_default);

#ifdef USE_AMAZON
	  try
	  {
	      Glib::RefPtr<Gdk::Pixbuf> cover;
	      Amazon::get_cover_by_asin (release.asin, cover);
	      merge_coverart->set (cover);
	  }
	  catch (...) {}
#endif

	  merge_album->set_text (release.title);
	  merge_artist->set_text (release.artist_name);
	  merge_artist_id->set_text (release.artist_id);
	  merge_ASIN->set_text (release.asin);
	  merge_release_ID->set_text (release.id);
	  merge_score->set_text ((boost::format ("%d") % release.score).str());

	  selected = release;

	  this->set_sensitive (true);
      }

      void
      LibraryUIMergeDialog::artist_changed ()
      {
	  using namespace Gtk;
	  using namespace MusicBrainzWS;

	  this->set_sensitive (false);
	
	  // Get releases from selected artist
	  TreeModel::const_iterator m_Iter = merge_cb_artist->get_active ();

	  ArtistReleaseMap::iterator rm_Iter = merge_artist_release_map.find ((*m_Iter)[artists.id]);
	  ReleaseIds &ids = rm_Iter->second;

	  merge_cb_album_store->clear ();
	  for (ReleaseIds::const_iterator r_Iter    = ids.begin (),
					  r_IterEnd = ids.end ();
					  r_Iter != r_IterEnd; ++r_Iter)
	  {
	      // Iter from releases map
	      ReleaseMap::iterator map_Iter = merge_releases.find (*r_Iter);

	      // Get the release based on id 
	      MB_Release const& release = map_Iter->second; 

	      TreeModel::iterator c_Iter = merge_cb_album_store->append ();
	      (*c_Iter)[albums.name] = release.title;
	      (*c_Iter)[albums.id] = release.id;
	  }

	  this->set_sensitive (true);
      }
}
