/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <config.h>
#include <string.h>

#include <glibmm.h>
#include <glib/gi18n.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#include <bmp/dbus.h>
#include <goa/libgoa.h>

#if 0
#ifdef HAVE_GUI
#  include <gtk/gtk.h>
#  include <chroma/chroma-list.h>
#  include <chroma/chroma-list-selection.h>
#  include <ui.hpp>
#  include <ui_util.hpp>
#endif
#endif

#include "file_utils.hpp"
#include "service_core.hpp"
#include "uri++.hpp"
#include "util.hpp"
#include "main.hpp"

#include <boost/algorithm/string.hpp>
#include <boost/shared_ptr.hpp>

namespace
{

  inline bool
  is_module (const std::string &basename)
  {
    return str_has_suffix_nocase (basename.c_str (), G_MODULE_SUFFIX);
  }

  enum
  {
    LIB_BASENAME,
    LIB_PLUGNAME,
    LIB_SUFFIX
  };

}

namespace Bmp
{

  struct CServiceCoreClass
  {
    GObjectClass parent;
  };

  namespace
  {
    GType	    c_service_core_type;
    void	    c_service_core_init (CServiceCoreClass *klass);
    void*	    c_service_core_parent_class = NULL;
    void	    c_service_core_intern_init (gpointer klass)
    {
      c_service_core_parent_class = g_type_class_peek_parent (klass);
      c_service_core_init ((CServiceCoreClass*) klass);
    }

    const GTypeInfo g_define_type_info = { 

        sizeof (CServiceCoreClass), 
        (GBaseInitFunc)		NULL, 
        (GBaseFinalizeFunc)	NULL, 
        (GClassInitFunc)	c_service_core_intern_init, 
        (GClassFinalizeFunc)	NULL, 
        NULL,  
        sizeof (CServiceCore), 
        0,      
        (GInstanceInitFunc)	NULL, 
    }; 

    GObject*
    c_service_core_constructor (GType                  type,
				guint                  n_construct_properties,
				GObjectConstructParam *construct_properties)
    {
	return reinterpret_cast<GObjectClass*>(c_service_core_parent_class)->constructor (type,
											   n_construct_properties,
											   construct_properties);
    }

    void
    c_service_core_init (CServiceCoreClass *klass)
    {
        GObjectClass *gobject_class = reinterpret_cast<GObjectClass*>(klass);
        gobject_class->constructor  = c_service_core_constructor;
    }


#define MPRIS_IDENTITY "BMP2"

  gboolean
  c_service_core_identity (CServiceCore     *CService,
			   char		   **identity,
			   GError	   **error)
  {
    *identity = g_strdup (MPRIS_IDENTITY);
    return TRUE;
  }

#if 0
  gboolean
  c_service_core_dummy	  (CServiceCore	    *CService,
			   GError	   **error)
  {
    CService->control->send_message (Bmp::ServiceCore::PREV);
    return TRUE;
  }
#endif

#include "service_core_glue.h"

  void initialize_dbus (CServiceCore *control)
  {
    DBusGConnection *bus;
    DBusGProxy	    *bus_proxy;
    gboolean	     connected = FALSE;
    guint	     request_name_result;
    GError	    *error = NULL;

    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (!bus)
        {
            g_message ("Running: %s (DBUS Error: '%s')", G_OBJECT_TYPE_NAME(control), error->message);
            g_error_free (error);
            return;
        }

    bus_proxy = dbus_g_proxy_new_for_name (bus,
					   "org.freedesktop.DBus",
					   "/org/freedesktop/DBus",
					   "org.freedesktop.DBus");

    if (!dbus_g_proxy_call (bus_proxy, "RequestName", &error,
                            G_TYPE_STRING, BMP_DBUS_INTERFACE, 
                            G_TYPE_UINT, 0,
                            G_TYPE_INVALID,
                            G_TYPE_UINT, &request_name_result,
                            G_TYPE_INVALID))
        {
            g_error ("%s: Failed RequestName request: %s", G_STRFUNC, error->message);
            g_error_free (error);
            error = NULL;
        }

    switch (request_name_result)
        {
        case DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER:
                {
                    dbus_g_connection_register_g_object (bus, BMP_DBUS_PATH, G_OBJECT(control));
                    connected = TRUE;
                }
            break;

        case DBUS_REQUEST_NAME_REPLY_EXISTS:
                {
                    connected = FALSE;
                }
            break;
        }

    if (connected)
        {
            g_message ("Running: %s (DBus:%s)", G_OBJECT_TYPE_NAME(control), BMP_DBUS_INTERFACE);
        }
    else
	{
            g_message ("Running: %s", G_OBJECT_TYPE_NAME(control));
	}
  }

} // ns:<unnamed>

    CServiceCore*
    c_service_core_new (Bmp::ServiceCore &control)
    {
      CServiceCore *CService;
  
      dbus_g_object_type_install_info (c_service_core_type, &dbus_glib_c_service_core_object_info); 
      CService = G_TYPE_CHECK_INSTANCE_CAST(g_object_new (c_service_core_type, NULL), c_service_core_type, CServiceCore);
      CService->control = &control;

      //Set up DBus connection
      initialize_dbus (CService); 
      return CService;
    }

    bool
    ServiceCore::load_plugins (const std::string& path)
    {
	using namespace boost::algorithm;

	std::string basename (Glib::path_get_basename (path));
	std::string pathname (Glib::path_get_dirname  (path));

	if (!is_module (basename)) return false;

	std::vector<std::string> subs;
	split (subs, basename, is_any_of ("_."));
	std::string dir   = Glib::build_filename (PLUGIN_DIR, "flow++"); 
	std::string name  = "flow" + std::string("_") + subs[LIB_PLUGNAME];
	std::string mpath = Glib::Module::build_path (dir, name);

	Glib::Module module (mpath, Glib::MODULE_BIND_LAZY);

	if (module)
	{
	    module.make_resident ();
	    Bmp::Plugin::FlowBase* (*plugin_create) ();
	    if (!module.get_symbol ("plugin_create", (void*&)plugin_create))
	    {
	      g_critical ("Unable to acquire plugin_create symbol from library %s", mpath.c_str ());
	    }
	    else
	    {
	      Bmp::Plugin::FlowBase* plugin = plugin_create ();
	      plugins.insert (std::make_pair (subs[LIB_PLUGNAME], plugin));
	      g_message ("Flow plugin '%s' loaded.", subs[LIB_PLUGNAME].c_str ()); 
	    }
	}
	else
	{
	    g_critical ("Failed to load Flow plugin '%s': %s", mpath.c_str (), module.get_last_error().c_str() );
	}

	return false;
    }

    //ctor
    ServiceCore::ServiceCore () 
		: Glib::ObjectBase (typeid(ServiceCore))
    {
	c_service_core_type = g_type_register_static (G_TYPE_OBJECT, "CServiceCore", &g_define_type_info, GTypeFlags(0)); 

	Util::dir_for_each_entry (Glib::build_filename (PLUGIN_DIR, "flow++"), 
                                  sigc::mem_fun (*this, &Bmp::ServiceCore::load_plugins));

	plugin = plugins.find ("linear")->second;

	//For now, we create only one playlist and you should leave the playlist manipulation stuff ALONE
	tracklist = new Bmp::Tracklist;

	g_message ("Running: ServiceCore");
    }

    void
    ServiceCore::init_dbus ()
    {
	c_service_core = reinterpret_cast<GObject*>(c_service_core_new (*this));
    }

    //dtor
    ServiceCore::~ServiceCore ()
    {
	g_object_unref (c_service_core);
    }

};
