//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <config.h>

#include <glibmm.h>
#include <gtkmm.h>
#include <libglademm.h>
#include <cstring>

#include <iostream>
#include <sstream>
#include <string>
#include <boost/format.hpp>

#include "util.hpp"
#include "uri++.hpp"
#include "database.hpp"
#include "library.hpp"
#include "network.hpp"
#include "paths.hpp"
#include "amazon.hpp"
#include "lyrics.hpp"
#include "file_utils.hpp"
#include "ui_util.hpp"

#include "trackinfo.hpp"
#include "x_library.hpp"

#include "main.hpp"

namespace Bmp
{

  TrackInfo *
  TrackInfo::create ()
  {
      const std::string path = DATA_DIR "/glade/trackinfo.glade";

      Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);

      TrackInfo *track_info = 0;
      glade_xml->get_widget_derived ("trackinfo", track_info);

      return track_info;
  }

  TrackInfo::TrackInfo (BaseObjectType                        *cobject,
                        const Glib::RefPtr<Gnome::Glade::Xml> &xml)
      : Gtk::Window (cobject),
        ref_xml (xml),
	tracker (new Bmp::Lyrictracker),
        path_image_default (DATA_DIR "/images/audio.png"),
        ok	(0),
        image	(0),
	lyrics	(0)
  {

      bmp_window_set_icon_list (GTK_WIDGET (this->gobj()), "player");

      ref_xml->get_widget ("ok", ok);
      ok->signal_clicked ().connect (sigc::mem_fun (this, &Bmp::TrackInfo::ok_clicked));

      ref_xml->get_widget ("image", image);
      image->set (path_image_default);

      ref_xml->get_widget ("lyrics", lyrics);

      ref_xml->get_widget ("entry_file",    entry_file);
      ref_xml->get_widget ("entry_artist",  entry_artist);
      ref_xml->get_widget ("entry_album",   entry_album);
      ref_xml->get_widget ("entry_title",   entry_title);
      ref_xml->get_widget ("entry_genre",   entry_genre);
      ref_xml->get_widget ("entry_comment", entry_comment);

      ref_xml->get_widget ("label_track",   label_track);
      ref_xml->get_widget ("label_date",    label_date);
      ref_xml->get_widget ("label_bitrate", label_bitrate);

      ref_xml->get_widget ("b_lyrics", b_lyrics);

      b_lyrics->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::TrackInfo::get_lyrics));
      b_lyrics->set_sensitive (Bmp::Network::is_connected());
  }

  void
  TrackInfo::ok_clicked ()
  {
      hide ();
  }

  void
  TrackInfo::get_lyrics ()
  {
      set_sensitive (false);

      Glib::ustring text;
      if (tracker->get_lyrics (track, text))
      {
	Glib::RefPtr<Gtk::TextBuffer> buffer = lyrics->get_buffer ();
	buffer->set_text (text);
	lyrics->set_sensitive (true);
	b_lyrics->set_sensitive (false);
      }

      set_sensitive (true);
  }

  void
  TrackInfo::clear ()
  {
      entry_file->set_text ("");
      entry_file->set_sensitive (false);

      entry_artist->set_text ("");
      entry_artist->set_sensitive (false);

      entry_album->set_text ("");
      entry_album->set_sensitive (false);

      entry_title->set_text ("");
      entry_title->set_sensitive (false);

      entry_genre->set_text ("");
      entry_genre->set_sensitive (false);

      entry_comment->set_text ("");
      entry_comment->set_sensitive (false);

      label_track->set_text ("");
      label_track->set_sensitive (false);

      label_date->set_text ("");
      label_date->set_sensitive (false);

      label_bitrate->set_text ("");
      label_bitrate->set_sensitive (false);

      Glib::RefPtr<Gtk::TextBuffer> buffer = lyrics->get_buffer ();
      buffer->set_text ("");
      lyrics->set_sensitive (false);
      b_lyrics->set_sensitive (Bmp::Network::is_connected());
      location.reset();

#ifdef USE_AMAZON
      image->set (path_image_default);
      image->drag_unhighlight ();
#endif

  }

  void
  TrackInfo::show (std::string uri)
  {
      Bmp::DB::DataRow row;
      try { library->get_metadata (uri, row); } catch (Bmp::Library::Exception& cxe) { return; }

      track = Bmp::Library::Track (row);

      if (location && !track.location.get().compare (location.get()))
      {
	Gtk::Window::show ();
	return;
      }
      else
      {
	location = track.location.get (); 
      }

      clear ();

      set_size_request (600, -1);
      set_keep_above (mcs->key_get<bool>("bmp","keep-above"));
      set_sensitive (false);
      Gtk::Window::show ();

      if (track.location)
      {
	  entry_file->set_text (track.location.get()); 
	  entry_file->set_sensitive (true);
      }

      if (track.artist)
      {
	  entry_artist->set_text (track.artist.get());
	  entry_artist->set_sensitive (true);
      }

      if (track.album)
      {
	  entry_album->set_text (track.album.get());
	  entry_album->set_sensitive (true);
      }

      if (track.title)
      {
	  entry_title->set_text (track.title.get());
	  entry_title->set_sensitive (true);
      }

      if (track.genre)
      {
	  entry_genre->set_text (track.genre.get());
	  entry_genre->set_sensitive (true);
      }

      if (track.comment)
      {
	  entry_comment->set_text (track.comment.get());
	  entry_comment->set_sensitive (true);
      }

      if (track.date)
      {
	  label_date->set_text ((boost::format ("%d") % track.date.get()).str ());
	  label_date->set_sensitive (true);
      }

      if (track.tracknumber)
      {
	  label_track->set_text ((boost::format ("%d") % track.tracknumber.get()).str ());
	  label_track->set_sensitive (true);
      } 

      if (track.bitrate)
      {
  	  label_bitrate->set_text ((boost::format ("%d") % track.bitrate.get()).str () + " kbit/s");
	  label_bitrate->set_sensitive (true);
      }

#ifdef USE_AMAZON
      Glib::RefPtr<Gdk::Pixbuf> cover;
      try {
	  Bmp::Amazon::get_cover (uri, cover);
	  image->set (cover);
	  image->drag_highlight ();
	  while (gtk_events_pending ()) gtk_main_iteration ();
      } catch (...) {}
#endif

      Glib::ustring text;
      if (tracker->get_lyrics (track, text, true))
      {
	Glib::RefPtr<Gtk::TextBuffer> buffer = lyrics->get_buffer ();
	buffer->set_text (text);
	lyrics->set_sensitive (true);
	if (Bmp::Network::is_connected ())
	{
	  b_lyrics->set_sensitive (false);
	}
      }

      set_sensitive (true);
  }

} // namespace Bmp
