#include <glibmm.h>

#include "database.hpp"
#include "library.hpp"
#include "uri++.hpp"
#include "file_utils.hpp"

#include "tracklist.hpp"
#include "main.hpp"

#include "x_library.hpp"
#include "x_playlist.hpp"
#include "x_play.hpp"
#include "x_playlist.hpp"
#include "x_vfs.hpp"

namespace Bmp
{
  // class History

  PlaybackHistory::PlaybackHistory ()
      : mark (history.begin ())
  {}

  PlaybackHistory::~PlaybackHistory ()
  {}

  void
  PlaybackHistory::rewind ()
  {
      mark = history.begin ();
  }

  void
  PlaybackHistory::set (GUID guid)
  {
      history.erase (mark, history.end ());
      history.push_back (guid);
  }

  void
  PlaybackHistory::append (GUID guid)
  {
      history.push_back (guid);
  }

  void
  PlaybackHistory::prepend (GUID guid)
  {
      history.push_front (guid);
  }

  void
  PlaybackHistory::back ()
  {
      --mark;
  }

  void
  PlaybackHistory::forward ()
  {
      ++mark;
  }

  void
  PlaybackHistory::clear ()
  {
      history.clear ();
      mark = history.begin ();
  }

  void
  PlaybackHistory::remove (GUID guid)
  {
      History::iterator item = std::find (history.begin (), history.end (), guid);
      history.erase (item);
  }

  // class Tracklist

  void
  Tracklist::row_inserted (const Gtk::TreeModel::Path     &path,
                           const Gtk::TreeModel::iterator &iter)
  {
      GUID guid = guid_base++;
      (*iter)[tracklist_columns.guid] = guid;
      mapping.insert (std::make_pair (guid, iter));
  }

  unsigned int
  Tracklist::insert_rows (const Util::FileList	&uris,
                          unsigned int		index)
  {
      using namespace Gtk;
      using namespace Bmp::DB;
      using namespace Bmp::Library;
      using namespace std;

      TreeModel::Path path;
      path.append_index (index);
      TreeModel::iterator iter_aux, iter_m;

      iter_aux = playlist->get_iter (path);
      if (iter_aux)
      {
	  iter_m = playlist->insert_after (iter_aux);
      }

      for (Util::FileList::const_iterator iter = uris.begin (), end = uris.end (); iter != end; ++iter)
      {
	  URI u (*iter);
	  URI::Protocol p = u.get_protocol ();
	  string uri = u; 
	  DataRow row;

	  if (!vfs->has_container (uri) || ((play->is_audio_file (uri) || (p == Bmp::URI::PROTOCOL_CDDA))))
	  {
	      try
	      {
		library->get_metadata (*iter, row);
		const Track track (row); 

		if (!iter_m)
		    iter_m = playlist->append ();
		else
		    iter_m = playlist->insert_after (iter_m); 

		(*iter_m)[tracklist_columns.uri] = Bmp::URI (*iter); 

		if (track.artist) (*iter_m)[tracklist_columns.artist] = track.artist.get();
		if (track.album) (*iter_m)[tracklist_columns.album] = track.album.get();
		if (track.title) (*iter_m)[tracklist_columns.title] = track.title.get();
		if (track.genre) (*iter_m)[tracklist_columns.genre] = track.genre.get();
		if (track.tracknumber) (*iter_m)[tracklist_columns.track] = track.tracknumber.get(); 
		if (track.date) (*iter_m)[tracklist_columns.date] = track.date.get(); 
		if (track.duration) (*iter_m)[tracklist_columns.time] = track.duration.get(); 

		++index;
	      }
	      catch (Bmp::Library::Exception cxe) { continue; }
	  }
	  else
	  {
	      VFS::Handle handle = VFS::Handle::create (uri);
	      Util::FileList list;

	      if (Glib::file_test (uri, Glib::FILE_TEST_IS_DIR))
		  Util::collect_path (uri, list);
	      else
		try { vfs->read (handle, list); } catch (...) { return index; }

	      for (Util::FileList::const_iterator i = list.begin (); i != list.end (); ++i)
	      {
		try
		{
		  library->get_metadata (*i, row);
		  const Track track (row); 

		  if (!iter_m)
		      iter_m = playlist->append ();
		  else
		      iter_m = playlist->insert_after (iter_m); 

		  (*iter_m)[tracklist_columns.uri] = Bmp::URI (*iter); 

		  if (track.artist) (*iter_m)[tracklist_columns.artist] = track.artist.get();
		  if (track.album) (*iter_m)[tracklist_columns.album] = track.album.get();
		  if (track.title) (*iter_m)[tracklist_columns.title] = track.title.get();
		  if (track.genre) (*iter_m)[tracklist_columns.genre] = track.genre.get();
		  if (track.tracknumber) (*iter_m)[tracklist_columns.track] = track.tracknumber.get(); 
		  if (track.date) (*iter_m)[tracklist_columns.date] = track.date.get(); 
		  if (track.duration) (*iter_m)[tracklist_columns.time] = track.duration.get(); 

		  ++index;
		}
		catch (Bmp::Library::Exception cxe) { continue; }
	      }
	  }
      }
      return index;
  }

#if 0
  std::string
  Tracklist::get_title_nth (int index)
  {
    using namespace Gtk;

    TreeModel::Path path;
    path.append_index (index);
    TreeModel::iterator i = playlist->get_iter (path); 

    Glib::ustring artist, album, title;

    artist = (*i)[tracklist_columns.artist];
    album  = (*i)[tracklist_columns.album];
    title  = (*i)[tracklist_columns.title];

    std::stringstream titlestring;
    titlestring << artist << " - " << album << " - " << title;
    return titlestring . str ();
  }
#endif

  void
  Tracklist::path_remove (GtkTreePath *path_c)
  {
      Gtk::TreePath path (path_c);      
      Gtk::TreeModel::iterator iter = playlist->get_iter (path);
      GUID guid = (*iter)[tracklist_columns.guid];
      mapping.erase (guid); 
  }

  Tracklist::Tracklist ()
  {
      playlist = Gtk::ListStore::create (tracklist_columns);
      playlist->signal_row_inserted ().connect (sigc::mem_fun (this, &Tracklist::row_inserted));
  }

  Tracklist::~Tracklist ()
  {}

  unsigned long Tracklist::guid_base = 0;

} // Bmp namespace

