#include <cstdlib>
#include <cstring>
#include <cmath>
#include <gtk/gtk.h>
#include <gtkmm.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>

#include <glib/gi18n.h>
#include <goa/libgoa.h>

#include "ui_util.hpp"
#include "util.hpp"

#include "main.hpp"
#include "paths.hpp"

#define _NET_WM_STATE_REMOVE   0
#define _NET_WM_STATE_ADD      1
#define _NET_WM_STATE_TOGGLE   2

#define _NET_WM_MOVERESIZE_SIZE_TOPLEFT      0
#define _NET_WM_MOVERESIZE_SIZE_TOP          1
#define _NET_WM_MOVERESIZE_SIZE_TOPRIGHT     2
#define _NET_WM_MOVERESIZE_SIZE_RIGHT        3
#define _NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT  4
#define _NET_WM_MOVERESIZE_SIZE_BOTTOM       5
#define _NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT   6
#define _NET_WM_MOVERESIZE_SIZE_LEFT         7
#define _NET_WM_MOVERESIZE_MOVE              8

using namespace Bmp;

namespace
{
    GQuark     quark_popup_data;
}

#define BMP_N_ICON_SIZES 5

void
bmp_window_set_icon_list (GtkWidget *window, const char *icon_name)
{
	static const char* sizes[] = {
		"16", "32", "48", "64", "128"
	};

	GdkPixbuf *pixbufs[BMP_N_ICON_SIZES];
	GError *error = NULL;
	GList *icon_list = NULL;
	const char *icon_theme;
	char *aux0, *aux1;
	int n;

	icon_theme = mcs->key_get<std::string>("bmp", "icon-theme"). c_str ();

	for (n = 0; n < BMP_N_ICON_SIZES; n++)
	  {
		aux0 = g_strconcat("icon_", icon_name, "_", sizes[n], ".png", NULL);
		aux1 = g_build_filename (DATA_DIR, BMP_BASENAME_ICON_DIR, icon_theme, aux0, NULL);

		pixbufs[n] = gdk_pixbuf_new_from_file (aux1, &error);

		if (error != NULL)
		  {
			g_message (error->message);
			g_error_free (error);
			error = NULL;
		  }

		icon_list = g_list_append (icon_list, pixbufs[n]);
		g_free (aux1);
		g_free (aux0);
	  }

	gtk_window_set_icon_list (GTK_WINDOW (window), icon_list);
	g_list_free (icon_list);

	for (n = 0; n < BMP_N_ICON_SIZES; n++)
		g_object_unref (pixbufs[n]);
}

#if 0
void
bmp_window_set_icon_list_pp (Gtk::Window     &window,
			     std::string     &icon_name)
{
	static const char* sizes[] = {
		"16", "32", "48", "64", "128"
	};

	Glib::ListHandle<Glib::RefPtr<Gdk::Pixbuf> > icons;
	Glib::ustring icon_theme;

	icon_theme = mcs->key_get<std::string>( "bmp", "icon-theme");

	for (unsigned int n = 0; n < BMP_N_ICON_SIZES; n++)
	  {
		Glib::RefPtr<Gdk::Pixbuf> pixbuf;
		std::string icon_filename;

		icon_filename = DATA_DIR +
				G_DIR_SEPARATOR_S +

				BMP_BASENAME_ICON_DIR +
				G_DIR_SEPARATOR_S +

				icon_theme +
				G_DIR_SEPARATOR_S +

				"icon_" + icon_name + "_" + sizes[n] + ".png";


		pixbuf = Gdk::Pixbuf::create_from_file (icon_filename);
		icons. push_back (pixbuf);
	  }

      window. set_icon_list (icons);
      icons. erase ();
}
#endif

static gdouble
mm_to_inch (gdouble mm)
{
    return mm / 25.40;
}

gdouble
screen_get_resolution (GdkScreen *screen)
{
  /* NOTE: this assumes the width and height resolutions are
     equal. This is only true for video modes with aspect ratios of
     4:3 e.g. 640x480, 800x600, etc. - descender */
  return std::ceil(screen_get_y_resolution (screen));
}

gdouble
screen_get_x_resolution (GdkScreen *screen)
{
  g_return_val_if_fail (GDK_IS_SCREEN (screen), -1.0);

  return (gdouble) gdk_screen_get_width (screen) / mm_to_inch (gdk_screen_get_height_mm (screen));
}

gdouble
screen_get_y_resolution (GdkScreen *screen)
{
  g_return_val_if_fail (GDK_IS_SCREEN (screen), -1.0);

  return (gdouble) gdk_screen_get_height (screen) / mm_to_inch (gdk_screen_get_height_mm (screen));
}

GtkWidget*
ui_manager_get_popup (GtkUIManager *ui_manager,
		      const char   *path)
{
    GtkWidget *menu_item;
    GtkWidget *submenu;

    menu_item = gtk_ui_manager_get_widget (ui_manager, path);

    if (GTK_IS_MENU_ITEM (menu_item))
	submenu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (menu_item));
    else
	submenu = NULL;

    return submenu;
}

static void
menu_popup_pos_func (GtkMenu * menu,
		     int * x,
		     int * y,
		     gboolean * push_in,
		     int * point)
{
    *x = point[0];
    *y = point[1];
    *push_in = FALSE;
}

void
bmp_menu_popup (GtkMenu * menu,
		int x,
		int y,
		unsigned int button,
		unsigned int time)
{
    int pos[2];

    pos[0] = x;
    pos[1] = y;

    gtk_menu_popup (menu, NULL, NULL,
		    (GtkMenuPositionFunc) menu_popup_pos_func, pos,
		    button, time);
}

static void
util_menu_delete_popup_data (GtkObject * object,
                             GtkItemFactory * ifactory)
{
    gtk_signal_disconnect_by_func (object,
				   GTK_SIGNAL_FUNC
				   (util_menu_delete_popup_data), ifactory);
    gtk_object_remove_data_by_id (GTK_OBJECT (ifactory), quark_popup_data);
}

#ifdef ENABLE_NLS
char *
bmp_menu_translate (const char *message,
                    gpointer     data)
{
  char *translation = (char *) Q_(message);

  if (!translation)
    {
      g_warning ("Bad translation for message: %s", message);
      translation = (char *) message;
    }

  return translation;
}
#endif


void
bmp_window_set_busy (GtkWidget *widget)
{
    GdkCursor *cursor;

    cursor = gdk_cursor_new_from_name (gdk_display_get_default (), "watch");
    if (!cursor) cursor = gdk_cursor_new_for_display (gdk_display_get_default (), GDK_WATCH);
    gdk_window_set_cursor (GTK_WIDGET(widget)->window, cursor);
}

void
bmp_window_set_idle (GtkWidget *widget)
{
    gdk_window_set_cursor (GTK_WIDGET(widget)->window, NULL);
}

namespace Bmp
{
  namespace Util
  {
    void
    window_set_busy (Gtk::Window &window)
    {
        window.get_window()->set_cursor (Gdk::Cursor (Gdk::Display::get_default(), "watch"));
    }

    void
    window_set_idle (Gtk::Window &window)
    {
        window.get_window()->set_cursor ();
    }

    // NOTE: gtkmm has no binding for gdk_screen_get_rgba_visual()
    Glib::RefPtr<Gdk::Visual>
    screen_get_rgba_visual (const Glib::RefPtr<Gdk::Screen> &screen)
    {
        return Glib::wrap (gdk_screen_get_rgba_visual (screen->gobj ()), true);
    }

    // NOTE: gtkmm has no binding for gdk_screen_get_rgba_colormap()
    Glib::RefPtr<Gdk::Colormap>
    screen_get_rgba_colormap (const Glib::RefPtr<Gdk::Screen> &screen)
    {
        return Glib::wrap (gdk_screen_get_rgba_colormap (screen->gobj ()), true);
    }

  } // namespace Util

} // namespace Bmp
