//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <cstring>
#include <glib/gi18n.h>
#include <glibmm.h>
#include <gtkmm.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>

#include <chroma/chroma-list.h>
#include <chroma/chroma-list-selection.h>
#include <chroma/chroma-vscrollbar.h>
#include <chroma/chroma-entry.h>
#include <goa/libgoa.h>
#include <boost/format.hpp>

#include "widgets/bmp_window.h"

#include "winamp2_skin.hpp"
#include "winamp2_main.hpp"

#include "action_defines.hpp"

#include "main.hpp"
#include "paths.hpp"
#include "ui_util.hpp"

#include "x_library.hpp"
#include "x_play.hpp"
#include "x_playlist.hpp"
#include "x_system_control.hpp"
#include "x_resource_manager.hpp"

#include "shell.hpp"

using namespace Bmp;

#include "gmacros.hpp"
BMP_DEFINE_TYPE (BmpWindowPlaylist, bmp_window_playlist, G_TYPE_OBJECT)

#define BMP_UIC_PL_INFORMATION	      "bmp_pct_pl_information"

#define BMP_UIC_PL_TITLE	      "bmp_pct_pl_title"
#define BMP_UIC_PL_TITLE_NOFOCUS      "bmp_pct_pl_title_nf"

#define BMP_UIC_PL_PAD_TOP	      "bmp_pct_pl_pad_top"
#define BMP_UIC_PL_PAD_TOP_NOFOCUS    "bmp_pct_pl_pad_top_nf"

#define BMP_UIC_PL_PAD_BOTTOM	      "bmp_pct_pl_pad_bottom"

#define BMP_UIC_PL_CORNER_UL	      "bmp_pct_pl_corner_ul"
#define BMP_UIC_PL_CORNER_UL_NOFOCUS  "bmp_pct_pl_corner_ul_nf"

#define BMP_UIC_PL_CORNER_UR	      "bmp_pct_pl_corner_ur"
#define BMP_UIC_PL_CORNER_UR_NOFOCUS  "bmp_pct_pl_corner_ur_nf"

#define BMP_UIC_PL_EDGE_LEFT	      "bmp_pct_pl_edge_left"
#define BMP_UIC_PL_EDGE_RIGHT	      "bmp_pct_pl_edge_right"

#define BMP_UIC_PL_CORNER_LL	      "bmp_pct_pl_corner_ll"
#define BMP_UIC_PL_CORNER_LR	      "bmp_pct_pl_corner_lr"

#define VSCROLLBAR_DEFAULT_WIDTH 14
#define PLAYLIST_INNER_XMARGIN 22
#define PLAYLIST_INNER_YMARGIN 42
#define PLAYLIST_ROOT_X 12
#define PLAYLIST_ROOT_Y 24
#define PLAYLIST_INNERPAD 2

#define REGION_L(x1,x2,y1,y2,w,h)                   \
    (event->x >= (x1) && event->x < (x2) &&     \
     event->y >= h - (y1) &&  \
     event->y < h - (y2))

#define REGION_R(x1,x2,y1,y2,w,h) \
    ((event->x >= (w-x1)) && \
     (event->x < (w-x2)) && \
     (event->y >= (h-y1)) && \
     (event->y < (h-y2)))

struct _BmpWindowPlaylistPrivate
{
    bool dispose_has_run;
    GHashTable *ui_components;
    int	columns_visible;
    bool focused;
};

GtkWidget *c_tracklist = 0;
GtkWidget *c_tracklist_scrollbar = 0;

namespace
{
  struct BmpPlaylistComponentT
  {
	gchar	   *name;
	guint	    x_origin;
	guint	    y_origin;
	guint	    width;
	guint	    height;
	GdkPixbuf  *pixbuf;
  };

  BmpPlaylistComponentT bmp_playlist_component_data[] = {

      {BMP_UIC_PL_TITLE,         26,  0, 100, 20 },
      {BMP_UIC_PL_TITLE_NOFOCUS,      26, 21, 100, 20 },

      {BMP_UIC_PL_PAD_TOP,      127,  0,  25, 20 },
      {BMP_UIC_PL_PAD_TOP_NOFOCUS,   127, 21,  25, 20 },

      {BMP_UIC_PL_PAD_BOTTOM,   179,  0,  25, 38 },

      {BMP_UIC_PL_CORNER_UL,      0,  0,  25, 20 },
      {BMP_UIC_PL_CORNER_UL_NOFOCUS,   0, 21,  25, 20 },

      {BMP_UIC_PL_CORNER_UR,    153,  0,  25, 20 },
      {BMP_UIC_PL_CORNER_UR_NOFOCUS, 153, 21,  25, 20 },

      {BMP_UIC_PL_EDGE_LEFT,      0, 42,  12, 29 },
      {BMP_UIC_PL_EDGE_RIGHT,    32, 42,  19, 29 },

      {BMP_UIC_PL_CORNER_LL,      0, 72, 125, 38 },
      {BMP_UIC_PL_CORNER_LR,    126, 72, 150, 38 },

      {BMP_UIC_PL_INFORMATION,  205,  0,  75, 32 }

  };

  void
  window_playlist_create (BmpWindowPlaylist *w_playlist);

  void
  window_playlist_dispose (GObject *gobject);

  void
  tracklist_set_action_states (BmpWindowPlaylist *w_playlist, GtkTreePath *path);

  void
  destroy_value (gpointer data)
  {
    BmpPlaylistComponentT *pct;
    pct = (BmpPlaylistComponentT *) data;
    if (pct)
    {
      if (pct->pixbuf) g_object_unref (pct->pixbuf);
      free(pct);
    }
  }

  int
  tracklist_compare_func
	        (GtkTreeModel *model,
                 GtkTreeIter  *a,
                 GtkTreeIter  *b,
                 gpointer      data)

  {
	int cmp_result;
	Bmp::Library::DatumDefine define = library->get_metadatum_define (vcolumns_get_item_nth (GPOINTER_TO_INT(data)));
	

	if (define.type == Bmp::DB::VALUE_TYPE_INT)
	  {
		int val_a, val_b;
		gtk_tree_model_get (model, a, GPOINTER_TO_INT(data), &val_a, -1);
		gtk_tree_model_get (model, b, GPOINTER_TO_INT(data), &val_b, -1);
		return val_a - val_b;
	  }
	else
	  {
		char  *text_a,
		      *text_b;

		char  *string_a,
		      *string_b;

		gtk_tree_model_get (model, a, GPOINTER_TO_INT(data), &text_a, -1);
		gtk_tree_model_get (model, b, GPOINTER_TO_INT(data), &text_b, -1);

		if ( (text_a != NULL) && (text_b == NULL)) return  1;
		if ( (text_a == NULL) && (text_b == NULL)) return  0;
		if ( (text_a == NULL) && (text_b != NULL)) return -1;

		string_a = g_utf8_casefold (text_a, std::strlen(text_a));
		string_b = g_utf8_casefold (text_b, std::strlen(text_b));

		if ( (string_a != NULL) && (string_b == NULL)) return  1;
		if ( (string_a == NULL) && (string_b == NULL)) return  0;
		if ( (string_a == NULL) && (string_b != NULL)) return -1;

		cmp_result = g_utf8_collate (string_a, string_b);

		free (string_a);
		free (string_b);
		free (text_a);
		free (text_b);

		return cmp_result;
	  }
  }

  void
  tracklist_cell_data_func   (ChromaList      *w_playlist,
                            GtkTreeModel    *model,
                            GtkTreeIter     *iter,
                            GValue          *value,
                            gpointer         data)
  {
    int   minutes, seconds;
    int   i_time; 
    char *s_time;

    gtk_tree_model_get (model, iter, 6, &i_time, -1);

    g_value_unset (value);

    minutes = i_time / 60;
    seconds = i_time % 60;

    g_value_init (value, G_TYPE_STRING);
    s_time = g_strdup_printf ("%d:%-2.2d", minutes, seconds);
    g_value_set_string (value, s_time);
    free (s_time);
  }

  BmpPlaylistComponentT*
  playlist_get_component(BmpWindowPlaylist *w_playlist, const gchar *key)
  {
    return static_cast<BmpPlaylistComponentT*>(g_hash_table_lookup(w_playlist->priv->ui_components, key));
  }

  void
  window_playlist_set_geometry_hints(BmpWindowPlaylist *w_playlist)
  {
    GtkWindow *window = GTK_WINDOW(w_playlist->window);
    GdkGeometry geometry;
    GdkWindowHints mask;

    geometry.width_inc = 25;
    geometry.height_inc = 29;

    /* make sure the min width/height are multiple values of the resize increments */
    geometry.min_width = ((int) (BMP_WINDOW_MAIN_WIDTH * 1 / geometry.width_inc)) * geometry.width_inc;
    geometry.max_width = 0xFFFF;
    geometry.base_width = 0;

    /* make sure the min width/height are multiple values of the resize increments */
    geometry.min_height = ((int) (BMP_WINDOW_MAIN_HEIGHT * 2 / geometry.height_inc)) * geometry.height_inc;
    geometry.max_height = 0xFFFF;
    geometry.base_height = 0;

    mask = GdkWindowHints (GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE | GDK_HINT_RESIZE_INC | GDK_HINT_BASE_SIZE);

    gtk_window_set_geometry_hints  (GTK_WINDOW(window),
                                    GTK_WIDGET(window),
                                    &geometry,
				    mask);

  }

  gboolean
  on_window_delete_event (GtkWidget *widget, GdkEvent *event, BmpWindowPlaylist *w_playlist)
  {
    Glib::RefPtr<Gtk::ToggleAction> action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
					    (actions_toggle->get_action (BMP_TOGGLE_ACTION_PL));
    action->set_active (false);
    return TRUE;
  }

  gboolean
  on_window_focus_in (GtkWidget *widget, GdkEventFocus *event, BmpWindowPlaylist *w_playlist)
  {
    w_playlist->priv->focused = true;
    gtk_widget_queue_draw (widget);
    return FALSE;
  }

  gboolean
  on_window_focus_out (GtkWidget *widget, GdkEventFocus *event, BmpWindowPlaylist *w_playlist)
  {
    w_playlist->priv->focused = false;
    gtk_widget_queue_draw (widget);
    return FALSE;
  }

  gboolean
  on_window_expose (GtkWidget *widget, GdkEventExpose *event, BmpWindowPlaylist *w_playlist)
  {
      BmpPlaylistComponentT *pct = 0;
      cairo_t	      *cr = 0;
      unsigned int     w_width = 0,
		       w_height = 0,
		       pad_ctr = 0,
		       i = 0;

      bool focus  = w_playlist->priv->focused; 
      w_width	  = widget->allocation.width;
      w_height	  = widget->allocation.height;

      cr = gdk_cairo_create (widget->window);

      pad_ctr =  (w_width - 150)/25;

      pct = playlist_get_component(w_playlist, focus ? BMP_UIC_PL_CORNER_UL : BMP_UIC_PL_CORNER_UL_NOFOCUS);
      bmp_draw_pixbuf_simple(cr, pct->pixbuf, 0, 0);

      pct = playlist_get_component(w_playlist, focus ? BMP_UIC_PL_CORNER_UR : BMP_UIC_PL_CORNER_UR_NOFOCUS);
      bmp_draw_pixbuf_simple(cr, pct->pixbuf, (w_width-pct->width), 0);

      pct = playlist_get_component(w_playlist, focus ? BMP_UIC_PL_TITLE : BMP_UIC_PL_TITLE_NOFOCUS);
      bmp_draw_pixbuf_simple(cr, pct->pixbuf, (w_width/2)-(50), 0);

      pct = playlist_get_component(w_playlist, BMP_UIC_PL_CORNER_LL);
      bmp_draw_pixbuf_simple(cr, pct->pixbuf, 0, (w_height-pct->height));

      pct = playlist_get_component(w_playlist, BMP_UIC_PL_CORNER_LR);
      bmp_draw_pixbuf_simple(cr, pct->pixbuf, (w_width-pct->width), (w_height-pct->height));

      pct = playlist_get_component(w_playlist, focus ? BMP_UIC_PL_PAD_TOP : BMP_UIC_PL_PAD_TOP_NOFOCUS);

      for (i = 0; i < pad_ctr / 2; i++)
	{

	    // Left of title
	    bmp_draw_pixbuf_simple(cr, pct->pixbuf, 25 + i * 25, 0);

	    // Right of title
	    bmp_draw_pixbuf_simple(cr, pct->pixbuf, (w_width+100)/2+i*25, 0);
	}

      if (pad_ctr & 1)
	{
	    /* Odd tile count, so one remaining to draw. Here we split
	     * it into two and draw half on either side of the title */

	    gdk_draw_pixbuf(widget->window,
			    NULL,
			    pct->pixbuf,
			    0,
			    0,
			    (w_width/2)+((pad_ctr/2)*25) +50,
			    0,
			    13,
			    -1,
			    GDK_RGB_DITHER_NONE,
			    0,
			    0);

	    gdk_draw_pixbuf(widget->window,
			    NULL,
			    pct->pixbuf,
			    0,
			    0,
			    ((pad_ctr / 2) * 25) + 25,
			    0,
			    13,
			    -1,
			    GDK_RGB_DITHER_NONE,
			    0,
			    0);
	  }

	// Frame sides
	for (i = 0; i < (w_height - (20 + 38)) / 29; i++)
	  {
		pct = playlist_get_component(w_playlist, BMP_UIC_PL_EDGE_LEFT);
		gdk_draw_pixbuf(widget->window,
			    NULL,
			    pct->pixbuf,
			    0,
			    0,
			    0,
			    20 + i *29,
			    -1,
			    -1,
			    GDK_RGB_DITHER_NONE,
			    0,
			    0);

		pct = playlist_get_component(w_playlist, BMP_UIC_PL_EDGE_RIGHT);
		gdk_draw_pixbuf(widget->window,
			    NULL,
			    pct->pixbuf,
			    0,
			    0,
			    w_width-19,
			    20 + i *29,
			    -1,
			    -1,
			    GDK_RGB_DITHER_NONE,
			    0,
			    0);

	  }

	pad_ctr =  (w_width - 275) / 25;
	pct = playlist_get_component(w_playlist, BMP_UIC_PL_PAD_BOTTOM);

	for (i = 0; i < pad_ctr+4; i++)
	  {
	    bmp_draw_pixbuf_simple(cr, pct->pixbuf, (125+i*25), (w_height-pct->height));
	  }
	bmp_draw_pixbuf_simple(cr, pct->pixbuf, (125+(pad_ctr+3)*25) + 5, (w_height-pct->height));
	cairo_destroy (cr);
	return FALSE;
  }

  void
  tracklist_has_sel (BmpWindowPlaylist *w_playlist, gboolean active)
  {
	actions->get_action (BMP_ACTION_DEL)->set_sensitive (active);
  }

  void
  tracklist_has_items (BmpWindowPlaylist *w_playlist, gboolean active)
  {
	actions->get_action (BMP_ACTION_DEL_ALL)->set_sensitive (active);
	actions->get_action (BMP_ACTION_PLAY)->set_sensitive (active);
  }

  void
  on_bmp_display_tracklist_numbers_changed
  (MCS_CB_DEFAULT_SIGNATURE, BmpWindowPlaylist *w_playlist)
  {
    chroma_list_set_use_numbers (CHROMA_LIST(c_tracklist), boost::get<bool>(value));
  }

  void
  on_bmp_use_custom_cursors_changed
  (MCS_CB_DEFAULT_SIGNATURE, BmpWindowPlaylist *w_playlist)
  {
    bmp_window_playlist_configure (w_playlist);
  }

  void
  on_bmp_font_changed
  (MCS_CB_DEFAULT_SIGNATURE, BmpWindowPlaylist *w_playlist)
  {
    chroma_list_set_font (CHROMA_LIST(c_tracklist), pango_font_description_from_string (boost::get<std::string>(value) . c_str ()));
#if 0
    GtkWidget *widget = GTK_WIDGET (g_object_get_data (G_OBJECT (w_playlist->window), "entry"));
    PangoFontDescription *pfd = pango_font_description_from_string (mcs->key_get<std::string>("bmp", "font") . c_str ());
    gtk_widget_modify_font (widget, pfd);
    pango_font_description_free (pfd);
    gtk_widget_queue_draw (w_playlist->window);
#endif
  }

  gboolean
  on_window_key_press (GtkWidget   *widget,
                       GdkEventKey *event,
			gpointer    user_data)
  {
    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    if (event->keyval == GDK_Escape)
      {
	actions->get_action (BMP_ACTION_ICONIFY)->activate ();
	return TRUE;
      }
    return FALSE;
  }

  gboolean
  on_window_button_press (GtkWidget	    *widget,
			  GdkEventButton    *event,
			  BmpWindowPlaylist *w_playlist)
  {
    GtkWidget	      *menu;
    int		       x,
		       y,
		       w_width,
		       w_height,
		       w_xpos,
		       w_ypos;

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    x = int (event->x);
    y = int (event->y);
    w_width = GTK_WIDGET (widget)->allocation.width;
    w_height = GTK_WIDGET (widget)->allocation.height;

    gtk_window_get_position (GTK_WINDOW (w_playlist->window), &w_xpos, &w_ypos);

    if ((event->button == 3) && (event->type == GDK_BUTTON_PRESS))
      {
	    GtkTreePath *path = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION(c_tracklist));
	    tracklist_set_action_states (w_playlist, path);
	    bool cond = ((path != NULL) && (!chroma_list_selection_multiple_selected (CHROMA_LIST_SELECTION (c_tracklist))));
	    gtk_tree_path_free (path);

	    actions->get_action (BMP_ACTION_FILEINFO)->set_sensitive (cond);
	    menu = ui_manager_get_popup (ui_manager->gobj(), "/popup-tracklist/menu-tracklist");
	    gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);
	    return TRUE;
      }
    else
    if ((event->button == 1) && (event->type == GDK_BUTTON_PRESS))
      {
	if (REGION_L(12, 37, 29, 11, w_width, w_height))
	{
	        //ADD button
		GtkRequisition req;
		menu = ui_manager_get_popup (ui_manager->gobj(), "/popup-tracklist/menu-add");
		gtk_widget_size_request (menu, &req);
		bmp_menu_popup (GTK_MENU (menu), (w_xpos+14), (w_ypos+w_height) - 8 - req.height, event->button, event->time);
		return TRUE;
	}
        else
	if (((w_width-event->x) <= 24) && ((w_height-event->y) <=24))
	{
	      gtk_window_begin_resize_drag (GTK_WINDOW (widget),
                                              GDK_WINDOW_EDGE_SOUTH_EAST,
                                              event->button,
                                              int (event->x_root),
                                              int (event->y_root),
                                              event->time);
	      return TRUE;

	}
	else
        if ((((y < 24) && (x < (w_width - 12)))) ||
	    ((event->x > 130) && (event->x < (w_width-40)) && (event->y > (w_height - PLAYLIST_INNER_YMARGIN))))
    	{
	      gdk_window_raise (widget->window);
	      gtk_window_begin_move_drag (GTK_WINDOW (widget),
					   event->button,
					   int (event->x_root),
					   int (event->y_root),
					   event->time);
	      return TRUE;
	}
	else
	if ((y < 16) && (x > (w_width - 12)))
	{
	      Glib::RefPtr<Gtk::ToggleAction> action = 
			  Glib::RefPtr<Gtk::ToggleAction>::cast_static (actions_toggle->get_action (BMP_TOGGLE_ACTION_PL));
	      action->set_active (false);
	      return TRUE;
	}
      }
    return FALSE;
  }

  void
  tracklist_set_action_states (BmpWindowPlaylist *w_playlist, GtkTreePath *path)
  {
	if ((!path) && (!chroma_list_selection_multiple_selected (CHROMA_LIST_SELECTION(c_tracklist))))
	  {
	    tracklist_has_sel (w_playlist, FALSE);
	  }
	else 
	  {
	    tracklist_has_sel (w_playlist, TRUE);
	  }

	tracklist_has_items (w_playlist, (chroma_list_get_length (CHROMA_LIST(c_tracklist)) == 0) ? FALSE : TRUE);
  }

  gboolean
  activate (GtkAction *action)
  {
    gtk_action_activate (action);
    return FALSE;
  }

  void
  on_tracklist_cursor_position (ChromaList	*list,
			      gint		 row,
			      BmpWindowPlaylist *w_playlist)
  {
    GtkTreePath *path;

    path = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION(list));
    tracklist_set_action_states (w_playlist, path);
    gtk_tree_path_free (path);

    if ((row == -1) || (!path)) 
      {
	actions->get_action (BMP_ACTION_FILEINFO_CLEAR)->activate ();
      }
    else
      {
	actions->get_action (BMP_ACTION_FILEINFO_IF_VISIBLE)->activate ();
      }

    bmp_system_control_update_title (bmp_system_control);
  }

  void
  on_tracklist_dnd (ChromaList	    *list,
		  gint		     index,
		  gchar		   **uri_list,
		  BmpWindowPlaylist *w_playlist)
  {
    int new_ctr;

    bmp_system_control_add_uri_list (bmp_system_control,
			(const gchar**) uri_list,
			index,
			FALSE,
			FALSE,
		        0,
			&new_ctr,
			NULL);
  }

  void
  on_tracklist_row_activated (ChromaList	  *list,
			      int	          index)
  {
    bmp_system_control_play_track (bmp_system_control, index, NULL);
  }

  void
  on_system_control_items_removed (BmpSystemControl	*system_control,
				   GArray		*removed_rows,
				   int			 n_removed_rows,
				   gpointer		 data) 
  {
    if (n_removed_rows == -1)
    {
	chroma_list_update (CHROMA_LIST (c_tracklist));
	chroma_list_scroll_to_offset (CHROMA_LIST (c_tracklist), 0);
	actions->get_action (BMP_ACTION_CLEAR_PLAYBACK_HISTORY)->set_sensitive (false);
    }
    actions->get_action (BMP_ACTION_FILEINFO_CLEAR)->activate ();
  }

  void
  on_system_control_track_change	  (BmpSystemControl	*system_control,
				   BmpWindowPlaylist	*w_playlist)
  {
    actions->get_action (BMP_ACTION_CLEAR_PLAYBACK_HISTORY)->set_sensitive (true);

    if (mcs->key_get<bool>("bmp", "follow-current-track"))
      {
	  gint	position,
		length,
	        display,
		scroll;

	  position =
	      g_object_get_int (bmp_system_control, "current_index");
	  display =
	      chroma_list_get_n_item_display (CHROMA_LIST(c_tracklist));
	  length =
	      chroma_list_get_length (CHROMA_LIST(c_tracklist));
	  scroll =
	      position - (display/2);

	  if (scroll > (length-display)) scroll = (length-display);
	  if (scroll < 0) scroll = 0;

	  chroma_list_scroll_to_offset (CHROMA_LIST(c_tracklist), scroll);
      }

    gtk_widget_queue_draw (GTK_WIDGET(c_tracklist));
  }

  void
  playlist_row_swap_func (GtkTreeModel	*model,
			gint		 pos_1,
		        gint		 pos_2)
  {
    GtkTreePath  *path_a, *path_b;
    GtkTreeIter   iter_a,  iter_b;

    path_a = gtk_tree_path_new_from_indices (pos_1, -1);
    path_b = gtk_tree_path_new_from_indices (pos_2, -1);
    gtk_tree_model_get_iter (model, &iter_a, path_a);
    gtk_tree_model_get_iter (model, &iter_b, path_b);
    gtk_tree_path_free (path_a);
    gtk_tree_path_free (path_b);

    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (model),
					  GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, 
					  GTK_SORT_ASCENDING);

    gtk_list_store_swap (GTK_LIST_STORE (model), &iter_a, &iter_b);
  }

  void
  on_tracklist_path_removing (ChromaList *list,
			      GtkTreePath *path,
			      BmpWindowPlaylist *w_playlist)
  {
    bmp_playlist_row_removing (bmp_playlist, path);
  }

  void
  window_playlist_create (BmpWindowPlaylist *w_playlist)
  {
    int			     last_visible;
    double		     sum = 0.0;

    w_playlist->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_decorated (GTK_WINDOW (w_playlist->window), false);

    gtk_widget_add_events (w_playlist->window,  GDK_FOCUS_CHANGE_MASK | GDK_BUTTON_MOTION_MASK |
                          GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                          GDK_SCROLL_MASK | GDK_VISIBILITY_NOTIFY_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK );

    window_playlist_set_geometry_hints (w_playlist);

    bmp_window_set_icon_list (w_playlist->window, "playlist");
    gtk_window_set_title (GTK_WINDOW (w_playlist->window), _("Tracklist - BMP"));
    gtk_window_set_role (GTK_WINDOW (w_playlist->window), "bmp::window-playlist");

    GtkWidget *alignment = gtk_alignment_new (0.5,0.5,1.0,1.0);
    GtkWidget *box = gtk_hbox_new (false, 2);
#if 0
    GtkWidget *vbox = gtk_vbox_new (false, 2);
#endif

    gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), PLAYLIST_ROOT_Y, PLAYLIST_INNER_YMARGIN,
							 PLAYLIST_ROOT_X, PLAYLIST_INNER_XMARGIN);

    //Tracklist 
    c_tracklist = chroma_list_new ();
    c_tracklist_scrollbar = chroma_vscrollbar_new ();
    g_object_set_data (G_OBJECT(c_tracklist_scrollbar), "list", c_tracklist);

    chroma_list_set_adjustment	(CHROMA_LIST (c_tracklist),
				 CHROMA_VSCROLLBAR (c_tracklist_scrollbar));

    chroma_list_set_model	(CHROMA_LIST (c_tracklist),
			         GTK_TREE_MODEL (bmp_playlist_get_playlist (bmp_playlist)),
				 ChromaListRowSwapFunc (playlist_row_swap_func));

    chroma_list_set_use_numbers (CHROMA_LIST (c_tracklist), mcs->key_get<bool>("bmp", "display-tracklist-numbers"));

    for (unsigned int n = 0; n < vcolumns_get_n_items(); n++)
      {
	    Bmp::Library::DatumDefine define = library->get_metadatum_define (vcolumns_get_item_nth(n));
            chroma_list_set_column_header (CHROMA_LIST(c_tracklist), n, define.title); 
            chroma_list_set_column_show_header (CHROMA_LIST(c_tracklist), n, TRUE);

            gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (bmp_playlist_get_playlist (bmp_playlist)),
					      n,
					      GtkTreeIterCompareFunc (tracklist_compare_func),
					      GINT_TO_POINTER (n),
					      NULL);

      }

    chroma_list_set_cell_data_func (CHROMA_LIST(c_tracklist), 6, tracklist_cell_data_func, NULL); 
    chroma_list_set_column_visible (CHROMA_LIST(c_tracklist), COLUMN_URI, FALSE);
    chroma_list_set_column_visible (CHROMA_LIST(c_tracklist), COLUMN_GUID, FALSE);
    chroma_list_set_column_xalign  (CHROMA_LIST(c_tracklist), 2, 1.0); //Track
    chroma_list_set_column_xalign  (CHROMA_LIST(c_tracklist), 5, 1.0); //Date
    chroma_list_set_column_xalign  (CHROMA_LIST(c_tracklist), 6, 1.0); //Time

    last_visible = 0;
    unsigned int n = 0;

    static boost::format f_visible ("column_%d_visible");
    static boost::format f_width ("column_%d_width");

    for (n = 0; n < vcolumns_get_n_items(); n++)
      {
	double width;
	bool visible;

	visible = mcs->key_get<bool>("tracklist", (f_visible % n).str());

	if (visible)
	  {
	    width = mcs->key_get<double>("tracklist", (f_width % n).str()); 
	    w_playlist->priv->columns_visible += 1;
	    sum += width;
	    last_visible = n;
	    chroma_list_set_column_width (CHROMA_LIST(c_tracklist), n, width);
	  }
	else
	  {
	    chroma_list_set_column_visible (CHROMA_LIST(c_tracklist), n, FALSE);
	    chroma_list_set_column_show_header (CHROMA_LIST(c_tracklist), n, FALSE);
	  }
      }

    if (sum != 1.0)
      {
	chroma_list_set_column_width (CHROMA_LIST(c_tracklist), n, 1. - sum);
      }

    gtk_box_pack_start (GTK_BOX (box), GTK_WIDGET (c_tracklist), TRUE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (box), GTK_WIDGET (c_tracklist_scrollbar), FALSE, FALSE, 0);

#if 0
    GtkWidget *entry = chroma_entry_new ();
    gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (box), TRUE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (entry), FALSE, FALSE, 0);
    g_object_set_data (G_OBJECT (w_playlist->window), "entry", gpointer(entry));
#endif

    gtk_container_add (GTK_CONTAINER (w_playlist->window), alignment);
    gtk_container_add (GTK_CONTAINER (alignment), box);
#if 0
    gtk_container_add (GTK_CONTAINER (alignment), vbox);
#endif

    gtk_widget_set_size_request (GTK_WIDGET(c_tracklist_scrollbar), VSCROLLBAR_DEFAULT_WIDTH, -1);

    g_object_connect (G_OBJECT(alignment),

		      "swapped-signal::configure-event",
		      G_CALLBACK(gtk_widget_queue_draw),
		      w_playlist->window,
  
		      NULL);


    g_object_connect (G_OBJECT(c_tracklist),

		      "signal::cursor-position",
		      G_CALLBACK(on_tracklist_cursor_position),
		      w_playlist,

		      "signal::row-activated",
		      G_CALLBACK(on_tracklist_row_activated),
		      w_playlist,

		      "signal::dnd",
		      G_CALLBACK(on_tracklist_dnd),
		      w_playlist,

		      "signal::path-removing",
		      G_CALLBACK(on_tracklist_path_removing),
		      w_playlist,

		      NULL);

    gtk_widget_set_app_paintable (w_playlist->window, true);

    g_object_connect (G_OBJECT(w_playlist->window),

			"signal::key-press-event",
			G_CALLBACK(on_window_key_press),
			w_playlist,

			"signal::button-press-event",
			G_CALLBACK(on_window_button_press),
			w_playlist,

			"signal::delete-event",
			G_CALLBACK(on_window_delete_event),
			w_playlist,

			"signal::expose_event",
			G_CALLBACK(on_window_expose),
			w_playlist,

			"signal::focus-in-event",
			G_CALLBACK(on_window_focus_in),
			w_playlist,

			"signal::focus-out-event",
			G_CALLBACK(on_window_focus_out),
			w_playlist,

			NULL);


    Glib::RefPtr<Gtk::ToggleAction> toggle_action = 
	      Glib::RefPtr<Gtk::ToggleAction>::cast_static (actions_toggle->get_action (BMP_TOGGLE_ACTION_FOLLOW_CURRENT));
    mcs_bind->bind_toggle_action (toggle_action, "bmp", "follow-current-track");
    
    mcs->subscribe ("Ui", "bmp", "font",
		    sigc::bind (sigc::ptr_fun (&on_bmp_font_changed),
				w_playlist));

    mcs->subscribe ("Ui", "bmp", "use-custom-cursors",
		    sigc::bind (sigc::ptr_fun(&on_bmp_use_custom_cursors_changed),
			        w_playlist));

    mcs->subscribe ("Ui", "bmp", "display-tracklist-numbers",
		    sigc::bind (sigc::ptr_fun(&on_bmp_display_tracklist_numbers_changed),
                                w_playlist));

    g_object_connect (G_OBJECT(bmp_system_control),

		      "signal::track-change",
		      G_CALLBACK (on_system_control_track_change),
		      w_playlist,

		      "signal::tracklist-items-removed",
		      G_CALLBACK (on_system_control_items_removed),
		      w_playlist,

		      "swapped-signal::tracklist-items-added",
		      G_CALLBACK (chroma_list_update),
		      c_tracklist,

		      NULL);	

    Gtk::Window *w = Glib::wrap (GTK_WINDOW (w_playlist->window));
    w->add_accel_group (ui_manager->get_accel_group());

    actions->get_action (BMP_ACTION_CLEAR_PLAYBACK_HISTORY)->set_sensitive (false);
    actions->get_action (BMP_ACTION_ADD_CD)->set_sensitive (play->property_has_cdda().get_value());
  }

  //GObject
  void
  bmp_window_playlist_init (BmpWindowPlaylist * w_playlist)
  {
    w_playlist->window = NULL;

    w_playlist->priv = g_new0 (BmpWindowPlaylistPrivate, 1);
    w_playlist->priv->dispose_has_run = FALSE;
    w_playlist->priv->ui_components = NULL;
    w_playlist->priv->columns_visible = 0;
    w_playlist->priv->focused = false;

    window_playlist_create (w_playlist);
  }

  GObject *
  window_playlist_constructor (GType type,
	                     guint n_construct_properties,
         	             GObjectConstructParam * construct_properties)
  {
    GObject *obj;
    BmpWindowPlaylistClass *klass;
    GObjectClass *parent_class;

    klass	  = BMP_WINDOW_PLAYLIST_CLASS (g_type_class_peek (BMP_TYPE_WINDOW_PLAYLIST));
    parent_class  = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
    obj		  = parent_class->constructor (type,
					       n_construct_properties,
					       construct_properties);
    return obj;
  }

  void
  window_playlist_dispose (GObject *gobject)
  {
    BmpWindowPlaylist *w_playlist = (BmpWindowPlaylist *) gobject;

    if (w_playlist->priv->dispose_has_run) return;
    w_playlist->priv->dispose_has_run = TRUE;

    for (unsigned int n = 0; n < vcolumns_get_n_items(); n++)
      {
	  gchar     *key_name;
	  gdouble    width;

	  width = chroma_list_get_column_width (CHROMA_LIST(c_tracklist), n);

	  key_name = g_strdup_printf ("column_%d_width", n);
	  mcs->key_set<double>("tracklist", key_name, width);
	  free (key_name);
      }

    chroma_list_set_model (CHROMA_LIST(c_tracklist), NULL, NULL);
    gtk_widget_destroy (c_tracklist);
    gtk_widget_destroy (c_tracklist_scrollbar);
    gtk_widget_destroy (w_playlist->window);
  }

  void
  bmp_window_playlist_class_init (BmpWindowPlaylistClass * klass)
  {
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->dispose = window_playlist_dispose;
    gobject_class->constructor = window_playlist_constructor;
  }

  void
  set_cursors (BmpWindowPlaylist *w_playlist)
  {
    GtkWidget *window;

    if (!w_playlist)
      return;

    window = w_playlist->window;

    if (!GTK_WIDGET_REALIZED(window))
      return;

    SET_CURSOR_FOR_WIDGET (CURSOR_PLAYLIST, GTK_WINDOW(window));
  }

}

void
bmp_window_playlist_column_set_active (BmpWindowPlaylist *w_playlist, gint column, gboolean _visible_arg)
{
      gchar     *key_name;
      gboolean   visible;

      key_name = g_strdup_printf ("column_%d_visible", column);
      visible = mcs->key_get<bool>("tracklist", key_name);
      free (key_name);

      if (visible == _visible_arg) return;

      key_name = g_strdup_printf ("column_%d_visible", column);
      mcs->key_set<bool>("tracklist", key_name, _visible_arg);
      free (key_name);

      chroma_list_set_column_visible (CHROMA_LIST(c_tracklist), column, _visible_arg);
      chroma_list_set_column_show_header (CHROMA_LIST(c_tracklist), column, _visible_arg);

      if (_visible_arg)
	{
	  w_playlist->priv->columns_visible++;
	}
      else
	{
	  w_playlist->priv->columns_visible--;
	}

      for (unsigned int n = 0; n < vcolumns_get_n_items(); n++)
	{
	    chroma_list_set_column_width (CHROMA_LIST(c_tracklist), n, 1./w_playlist->priv->columns_visible);
            key_name = g_strdup_printf ("column_%d_width", n);
	    mcs->key_set<double>("tracklist", key_name, 1./w_playlist->priv->columns_visible);
	    free (key_name);
	}

    chroma_list_refresh (CHROMA_LIST(c_tracklist));
}

void
bmp_window_playlist_show_trackinfo (BmpWindowPlaylist *w_playlist)
{
    GtkListStore	  *list;
    GtkTreeRowReference	  *reference;
    GtkTreePath		  *path;

    list = bmp_playlist_get_playlist (bmp_playlist);
    reference = static_cast<GtkTreeRowReference *> (g_object_get_data (G_OBJECT(list), "current"));
    if (!reference) return;

    path = gtk_tree_row_reference_get_path (reference);
    chroma_list_set_cursor_position (CHROMA_LIST(c_tracklist), gtk_tree_path_get_indices (path)[0]);
    gtk_tree_path_free (path);

    actions->get_action (BMP_ACTION_FILEINFO)->activate ();
}

void
bmp_window_playlist_configure (BmpWindowPlaylist *w_playlist)
{
	GtkWidget *window = w_playlist->window;
	BmpPlaylistComponentT *pct = NULL;

	if (w_playlist->priv->ui_components != NULL)
	  {
	    g_hash_table_destroy (w_playlist->priv->ui_components);
	  }

        w_playlist->priv->ui_components = g_hash_table_new_full  (g_str_hash,
                                                            g_str_equal,
                                                            NULL,
                                                            destroy_value);

        for (unsigned int i = 0; i < G_N_ELEMENTS (bmp_playlist_component_data); i++)
        {
            pct = g_new0 (BmpPlaylistComponentT, 1);

            pct->name = 0;
            pct->pixbuf = bmp_pixbuf_new_simple (bmp_playlist_component_data[i].width,
                                                 bmp_playlist_component_data[i].height);
            bmp_pixbuf_copy_area_simple (pct->pixbuf,
                                         bmp_playlist_component_data[i].x_origin,
                                         bmp_playlist_component_data[i].y_origin,
                                         bmp_playlist_component_data[i].width,
                                         bmp_playlist_component_data[i].height);


            pct->x_origin = bmp_playlist_component_data[i].x_origin;
            pct->y_origin = bmp_playlist_component_data[i].y_origin;
            pct->width = bmp_playlist_component_data[i].width;
            pct->height = bmp_playlist_component_data[i].height;

            g_hash_table_insert	(w_playlist->priv->ui_components,
                                 bmp_playlist_component_data[i].name,
                                 pct);

	}

	GdkColor *bg_normal  = bmp_colors_get_color (BMP_SKINCOLOR_PL_BG_NORMAL);
	GdkColor *fg_normal  = bmp_colors_get_color (BMP_SKINCOLOR_PL_FG_NORMAL);
	GdkColor *fg_current = bmp_colors_get_color (BMP_SKINCOLOR_PL_FG_CURRENT);

#if 0
	GtkWidget *widget = GTK_WIDGET (g_object_get_data (G_OBJECT (w_playlist->window), "entry"));
	PangoFontDescription *pfd = pango_font_description_from_string (mcs->key_get<std::string>("bmp", "font") . c_str ());
	gtk_widget_modify_font (widget, pfd);
	pango_font_description_free (pfd);

        gtk_widget_modify_base (GTK_WIDGET (widget), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (widget), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (widget), GTK_STATE_NORMAL, fg_normal); 

        gtk_widget_modify_base (GTK_WIDGET (widget), GTK_STATE_ACTIVE, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (widget), GTK_STATE_ACTIVE, bg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (widget), GTK_STATE_ACTIVE, fg_normal); 

        gtk_widget_modify_base (GTK_WIDGET (widget), GTK_STATE_SELECTED, fg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (widget), GTK_STATE_SELECTED, fg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (widget), GTK_STATE_SELECTED, bg_normal); 

        gtk_widget_modify_text (GTK_WIDGET (widget), GTK_STATE_ACTIVE, fg_normal); 
        gtk_widget_modify_text (GTK_WIDGET (widget), GTK_STATE_SELECTED, bg_normal); 
#endif

        gtk_widget_modify_base (GTK_WIDGET (window), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (window), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (window), GTK_STATE_NORMAL, fg_normal); 

        gtk_widget_modify_base (GTK_WIDGET (c_tracklist), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (c_tracklist), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (c_tracklist_scrollbar), GTK_STATE_NORMAL, bg_normal); 

        gtk_widget_modify_fg (GTK_WIDGET (c_tracklist), GTK_STATE_NORMAL, fg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (c_tracklist_scrollbar), GTK_STATE_NORMAL, fg_normal); 

	gtk_widget_modify_fg (GTK_WIDGET (c_tracklist), GTK_STATE_PRELIGHT, fg_current); 
	gtk_widget_modify_fg (GTK_WIDGET (c_tracklist_scrollbar), GTK_STATE_PRELIGHT, fg_current); 

	set_cursors (w_playlist);

	gtk_widget_queue_draw (GTK_WIDGET (window));
}

BmpWindowPlaylist*
bmp_window_playlist_new (void)
{
  return BMP_WINDOW_PLAYLIST (g_object_new (BMP_TYPE_WINDOW_PLAYLIST, NULL)); 
}
