/*  BMPx - Cross-platform multimedia player
 *  Copyright (C) 2003-2004  Edward Brocklesby, Chong Kai Xiong,
 *            Milosz Derezynski, David Lau, Michiel Sikkes
 *
 *  ui.c/bmp_winamp2_skin.h
 *  bmp_button.c/bmp_button.h
 *  bmp_button_toggle.c/bmp_button_toggle.h
 *  bmp_slider.c/bmp_slider.h
 *
 *  M.Derezynski
 *  Tim-Philipp Mueller
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include <cstring>
#include <iostream>
#include <sstream>

#include "widgets/bmp_button.h"
#include "widgets/bmp_button_toggle.h"
#include "widgets/bmp_slider.h"
#include "widgets/bmp_window.h"
#include "widgets/bmp_status_icon.h"
#include <chroma/chroma-list.h>
#include <goa/libgoa.h>

#include "util.hpp"

#include "winamp2_playlist.hpp"
#include "winamp2_main.hpp"
#include "winamp2_skin.hpp"

//Utility/auxilliary stuff
#include "ui_util.hpp"
#include "main.hpp"
#include "paths.hpp"
#include "debug.hpp"

#include "x_library.hpp"
#include "x_play.hpp"
#include "x_system_control.hpp"
#include "x_resource_manager.hpp"

using namespace Bmp;

#include "gmacros.hpp"
BMP_DEFINE_TYPE (BmpWinamp2Skin, bmp_winamp2_skin, G_TYPE_OBJECT)

namespace
{
    bool dispose_has_run = false;
}

namespace 
{
  char      *bmp_skin_colors[BMP_N_SKINCOLORS];
  GdkColor  *bmp_skin_colors_gdk[BMP_N_SKINCOLORS];
  GdkPixbuf *bmp_cursors[CURSOR_N_CURSORS];
  std::vector<std::string> bmp_cursor_names;
  std::vector<std::string> bmp_skin_color_names;

  char* bmp_skin_files[] = {
        "main.bmp",
        "shufrep.bmp",
        "cbuttons.bmp",
        "posbar.bmp",
        "volume.bmp",
        "titlebar.bmp",
        "balance.bmp",
        "pledit.bmp",
        "monoster.bmp",
        NULL,
        "playpaus.bmp"
  };

  BmpSkinComponent bmp_skin_files_sequential[] = {
        SKIN_COMPONENT_MAIN,
        SKIN_COMPONENT_SHUFREP,
        SKIN_COMPONENT_CBUTTONS,
        SKIN_COMPONENT_TITLEBAR,
        SKIN_COMPONENT_PLEDIT,
        SKIN_COMPONENT_MONOSTER,
        SKIN_COMPONENT_PLAYPAUSE
  };
}

/* 'extern' referenced */
BmpPosbarT	  *bmp_posbar = 0;
BmpVolumeT	  *bmp_volume = 0;
#if 0
BmpVolumeT	  *bmp_balance = 0;
#endif
BmpSkinComponentT *bmp_skin_component[N_SKIN_COMPONENTS];
GdkBitmap	  *mask_window_main = 0;

namespace
{

#define PLEDIT_TEXT_SECTION_NAME "Text"

  GdkBitmap*

  create_transparent_mask

  (const std::string& path,
   const std::string& file,
   const std::string& section,
   int width,
   int height)

  {
	GdkBitmap	  *mask = NULL;
	GdkGC		  *gc = NULL;
	GdkPoint	  *points;
	GdkColor	   pattern;
	GArray		  *num,
			  *point;
	std::string	   filename;
	bool		  created_mask = false;

	if (!path.empty ())
	  {
	    Util::find_file (path, file, filename, 1);
	  }

	if (filename.empty ())
	  {
	    return NULL;
	  }

	if ((num   = (GArray*)read_ini_array (filename.c_str(), section.c_str(), "NumPoints")) == NULL)
	  {
	    return NULL;
	  }

	if ((point = (GArray*)read_ini_array (filename.c_str(), section.c_str(), "PointList")) == NULL)
	  {
	    g_array_free (num, TRUE);
	    return NULL;
	  }

	mask = gdk_pixmap_new (NULL, width, height, 1);
	gc = gdk_gc_new (mask);

	pattern.pixel = 0;
	gdk_gc_set_foreground (gc, &pattern);
	gdk_draw_rectangle (mask, gc, TRUE, 0, 0, width, height);
	pattern.pixel = 1;
	gdk_gc_set_foreground (gc, &pattern);

	int k;
	int j = 0;

	for (unsigned int i = 0; i < num->len; i++)
	{
	    if ((point->len - j) >= (g_array_index(num, gint, i) * 2))
	    {
		points = g_new (GdkPoint, g_array_index(num, gint, i));
		for (k = 0; k < g_array_index(num, gint, i); k++)
		{
		    points[k].x = g_array_index(point, gint, j + k * 2);
		    points[k].y = g_array_index(point, gint, j + k * 2 + 1);
		}
		j += k * 2;
		gdk_draw_polygon (mask, gc, TRUE, points, g_array_index(num, gint, i));
		g_free (points);
	    }
	    created_mask = true;
	}

	g_array_free (num, TRUE);
	g_array_free (point, TRUE);

	if (!created_mask)
	  {
	    gdk_draw_rectangle (mask, gc, TRUE, 0, 0, width, height);
	  }

	gdk_gc_destroy (gc);

	return mask;
  }

  GdkPixbuf*
  get_cursor_pixbuf
  (const std::string& archive_dirname,
   const std::string& cursor_filename)
  {
      std::string filename;
      if (!Util::find_file (archive_dirname, cursor_filename, filename, 1))
	    return NULL;

      GdkPixbufAnimation *cursor_animated =
	    gdk_pixbuf_animation_new_from_file (filename.c_str(), NULL);

      GdkPixbuf *pb_aux =
	    gdk_pixbuf_animation_get_static_image (cursor_animated);

      GdkPixbuf *pb_return =
	    gdk_pixbuf_copy (pb_aux);

      g_object_unref (cursor_animated);
      return pb_return;
  }

  unsigned int
  hex_chars_to_int(gchar hi, gchar lo)
  {
    /*
     * Converts a value in the range 0x00-0xFF
     * to a integer in the range 0-65535
     */
    gchar str[3];
    unsigned int value;

    str[0] = hi;
    str[1] = lo;
    str[2] = 0;

    value = (CLAMP(strtol(str, NULL, 16), 0, 0xFF) << 8);

    return value;
  }

  void
  skin_components_free (BmpWinamp2Skin *ui)
  {
      int n;

      for (n = 0; n < N_SKIN_COMPONENTS; n++)
      {
	 if (bmp_skin_component[n] != NULL)
	 {
	    g_object_unref (bmp_skin_component[n]->pixbuf);
	    g_free (bmp_skin_component[n]);
	    bmp_skin_component[n] = NULL;
	 }
      }

      if (bmp_posbar)
	{
	    g_object_unref (bmp_posbar->pixbuf);
	    g_free (bmp_posbar);
	    bmp_posbar = NULL;
	}

      if (bmp_volume)
	{
	    g_object_unref (bmp_volume->pixbuf);
	    g_free (bmp_volume);
	    bmp_volume = NULL;
	}

#if 0
      if (bmp_balance)
	{
	    g_object_unref (bmp_balance->pixbuf);
	    g_free (bmp_balance);
	    bmp_balance = NULL;
	}
#endif

  }

  void
  load_skin_components (BmpWinamp2Skin             *ui,
                        const std::string	   &archive_filename,
                        int			    type)
  {
      GdkPixbuf   *pixbuf;
      std::string filename;
      std::string archive_dirname;

      debug ("winamp2-skin", "Initializing Path.");
      archive_dirname = (type == BMP_SKIN_ARCHIVE) ? Util::archive_decompress (archive_filename) : archive_filename;

      debug ("winamp2-skin", "Freeing Components.");
      bmp_cursors_free ();
      skin_components_free (ui);

      debug ("winamp2-skin", "Loading Cursors from archive.");
      bmp_cursors_set (archive_dirname);

      // Buttons, Titlebars
      for (unsigned int n = 0; n < G_N_ELEMENTS(bmp_skin_files_sequential); n++)
      {
	  debug ("winamp2-skin", "Searching for: %s", bmp_skin_files[bmp_skin_files_sequential[n]]);
	  if (!Util::find_file (archive_dirname, bmp_skin_files[bmp_skin_files_sequential[n]], filename, 1))
	  {
		  g_warning ("%s not found", bmp_skin_files[bmp_skin_files_sequential[n]]);
	  }
	  else
	  {
		  pixbuf = gdk_pixbuf_new_from_file (filename.c_str(), NULL);
		  bmp_skin_component[bmp_skin_files_sequential[n]] = g_new0 (BmpSkinComponentT, 1);
		  bmp_skin_component[bmp_skin_files_sequential[n]]->identifier = bmp_skin_files_sequential[n];
		  bmp_skin_component[bmp_skin_files_sequential[n]]->pixbuf = pixbuf;
	  }
      }

      //Numbers
      debug ("winamp2-skin", "Loading numbers,");
      Util::find_file (archive_dirname, "numbers.bmp", filename, 1);
      if (filename.empty ())
	{
	    Util::find_file (archive_dirname, "nums_ex.bmp", filename, 1);
	}

      if (filename.empty ())
	{
	    g_warning ("%s: Numbers not found", G_STRFUNC);
	}
      else
	{
	    bmp_skin_component[SKIN_COMPONENT_NUMBERS] = g_new0 (BmpSkinComponentT, 1);

	    pixbuf = gdk_pixbuf_new_from_file (filename.c_str(), NULL);
	    bmp_skin_component[SKIN_COMPONENT_NUMBERS]->identifier = SKIN_COMPONENT_NUMBERS;
	    bmp_skin_component[SKIN_COMPONENT_NUMBERS]->pixbuf = pixbuf;

	    if (gdk_pixbuf_get_width(pixbuf) >= 108)
	      {
		  bmp_skin_component[SKIN_COMPONENT_NUMBERS]->user_data = GINT_TO_POINTER (12);
	      }
	    else
	      {
		  bmp_skin_component[SKIN_COMPONENT_NUMBERS]->user_data = GINT_TO_POINTER (10);
	      }
	}

      //Volume, Balance and Seek slider
      debug ("winamp2-skin", "Loading seek slider,");
      Util::find_file (archive_dirname, bmp_skin_files[SKIN_COMPONENT_POSBAR], filename, 1);
      if (filename.empty ())
	{
	    g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_POSBAR]);
	}
      else
       {
	    pixbuf = gdk_pixbuf_new_from_file (filename.c_str(), NULL);
	    bmp_posbar = g_new0 ( BmpPosbarT, 1 );
	    bmp_posbar->pixbuf = pixbuf;
       }

      debug ("winamp2-skin", "Loading volume slider,");
      Util::find_file (archive_dirname, bmp_skin_files[SKIN_COMPONENT_VOLUME], filename, 1);
      if (filename.empty ())
	{
	    g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_VOLUME]);
	}
      else
	{
	    pixbuf = gdk_pixbuf_new_from_file(filename.c_str(), NULL);
	    bmp_volume = g_new0 ( BmpVolumeT, 1 );
	    bmp_volume->pixbuf = pixbuf;
	    bmp_volume->height = gdk_pixbuf_get_height(pixbuf) - 422;

	    if (gdk_pixbuf_get_height(pixbuf) > 422)
		bmp_volume->has_button = TRUE;
	    else
		bmp_volume->has_button = FALSE;
	}

#if 0
      Util::find_file (archive_dirname, bmp_skin_files[SKIN_COMPONENT_BALANCE], filename, 1);
      if (filename.empty ())
	{
	    g_message ("%s not found", bmp_skin_files[SKIN_COMPONENT_BALANCE]);
	}
      else
	{
	    pixbuf = gdk_pixbuf_new_from_file(filename.c_str(), NULL);
	    bmp_balance = g_new0 ( BmpVolumeT, 1 );
	    bmp_balance->pixbuf = pixbuf;
	    bmp_balance->height = gdk_pixbuf_get_height(pixbuf) - 421;

	    if (gdk_pixbuf_get_height(pixbuf) > 421)
		bmp_balance->has_button = TRUE;
	    else
		bmp_balance->has_button = FALSE;
	}
#endif

      debug ("winamp2-skin", "Initializing colors.");
      bmp_colors_init (archive_dirname);

      debug ("winamp2-skin", "Initializing main window xshape mask");
      mask_window_main = create_transparent_mask (archive_dirname,
						  BMP_FILE_REGION,
						  BMP_MASK_REGION_SECTION_MAIN_WINDOW,
						  275,
						  116);

      if (type == BMP_SKIN_ARCHIVE)
	{
	    debug ("winamp2-skin", "Deleting directory.");
	    Util::del_directory (archive_dirname);
	}
  }

  void
  bmp_winamp2_skin_dispose (GObject *obj)
  {
    BmpWinamp2Skin *ui = BMP_WINAMP2_SKIN (obj);
    skin_components_free (ui);
    g_message("stopped: %s", G_OBJECT_TYPE_NAME(ui));
  }

  void
  bmp_winamp2_skin_class_init (BmpWinamp2SkinClass * klass)
  {
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose  = bmp_winamp2_skin_dispose;
  }

  void
  bmp_winamp2_skin_init (BmpWinamp2Skin *ui)
  {
    for (unsigned int n = 0; n < N_SKIN_COMPONENTS; bmp_skin_component[n++] = NULL);
    bmp_cursors_init ();
    mask_window_main = NULL;
  }
}

BmpWinamp2Skin*
bmp_winamp2_skin_new (void)
{
    BmpWinamp2Skin *ui;
    GtkSettings *settings;
    const char *name;
    ui = BMP_WINAMP2_SKIN (g_object_new (BMP_TYPE_WINAMP2_SKIN, NULL));

    // Hack-around for no xsettings manager available that would set the cursor theme
    settings = gtk_settings_get_for_screen (gdk_screen_get_default());
    name = g_object_get_string (settings, "gtk-cursor-theme-name");
    if (!name)
    {
      name = g_getenv("XCURSOR_THEME");
      g_object_set (settings, "gtk-cursor-theme-name", name, NULL);
    }
    return ui;
}

bool
bmp_winamp2_skin_load_skin (BmpWinamp2Skin     *ui,
			    const std::string&  filename)
{
  if (!Glib::file_test(filename, Glib::FILE_TEST_EXISTS))
    {
	return false;
    }

  bmp_colors_free();

  if (Glib::file_test(filename, Glib::FILE_TEST_IS_DIR)) 
    {
	load_skin_components (ui, filename, BMP_SKIN_DIR);
    }
  else if (Glib::file_test(filename, Glib::FILE_TEST_IS_REGULAR))
    {
	load_skin_components (ui, filename, BMP_SKIN_ARCHIVE);
    }
  else
    {
	return false;
    }

  return true; 
}

GdkPixbuf*
bmp_pixbuf_new_simple (unsigned int width, unsigned int height)
{
	GdkPixbuf *pixbuf;

	pixbuf = bmp_skin_component[SKIN_COMPONENT_PLEDIT]->pixbuf;
        return gdk_pixbuf_new (gdk_pixbuf_get_colorspace(pixbuf),
                         gdk_pixbuf_get_has_alpha(pixbuf),
                         gdk_pixbuf_get_bits_per_sample(pixbuf),
                         width, height);
}

void
bmp_pixbuf_copy_area_simple(GdkPixbuf *dest, unsigned int x, unsigned int y, unsigned int width, unsigned int height)
{
	GdkPixbuf * pixbuf;

	pixbuf = bmp_skin_component[SKIN_COMPONENT_PLEDIT]->pixbuf;
	gdk_pixbuf_copy_area (pixbuf, x, y, width, height, dest, 0, 0);
}

void
bmp_draw_pixbuf_simple (cairo_t *cr, GdkPixbuf *buf, unsigned int x, unsigned int y)
{
    cairo_save (cr);
    gdk_cairo_set_source_pixbuf (cr,
				 buf,
				 x,
				 y);
    cairo_paint (cr);
    cairo_restore (cr);
}

void
bmp_cursors_init ()
{
  const char *names[] = {
   "normal.cur",
   "posbar.cur",
   "volbar.cur",
   "pnormal.cur",
   "psize.cur",
   "mainmenu.cur"
  };

  for (unsigned int i = 0; i < G_N_ELEMENTS(names); i++)
    {
      bmp_cursors[i] = NULL;
      bmp_cursor_names.push_back (names[i]);
    }
}

void
bmp_cursors_free ()
{
  gint n;

  if (bmp_cursors[0] != NULL)
	for (n = 0; n < CURSOR_N_CURSORS; n++)
	  {
	     if (bmp_cursors[n] != NULL)
	       {
        	 g_object_unref (bmp_cursors[n]);
	       }
	     bmp_cursors[n] = NULL;
	  }
}

void
bmp_cursors_set (const std::string& archive_dirname)
{
  bmp_cursor_add (archive_dirname, CURSOR_NORMAL);
  bmp_cursor_add (archive_dirname, CURSOR_POSBAR);
  bmp_cursor_add (archive_dirname, CURSOR_VOLBAR);
  bmp_cursor_add (archive_dirname, CURSOR_PLAYLIST);
  bmp_cursor_add (archive_dirname, CURSOR_PLAYLIST_SIZE);
  bmp_cursor_add (archive_dirname, CURSOR_MAINMENU);
}

void
bmp_cursor_add (const std::string& archive_dirname,
	        int		   cursor_enum)
{
  bmp_cursors[cursor_enum] = get_cursor_pixbuf (archive_dirname.c_str(), bmp_cursor_names[cursor_enum].c_str());
}

GdkCursor*
bmp_cursor_get (::CursorId cursor_enum)
{
  GdkCursor * cursor;
  GdkPixbuf * pixbuf;
  const gchar *hot_x = NULL, *hot_y = NULL;
  gint hot_x_int = 0, hot_y_int = 0;

  pixbuf = bmp_cursors[cursor_enum];

  if (!mcs->key_get<bool>("bmp","use-custom-cursors")) return NULL;

  if (!pixbuf)
    {
      pixbuf = bmp_cursors[CURSOR_NORMAL];
      if (!pixbuf) return NULL;
    }

  hot_x = gdk_pixbuf_get_option(pixbuf,"x_hot");
  hot_y = gdk_pixbuf_get_option(pixbuf,"y_hot");

  if (hot_x) hot_x_int = std::atoi (hot_x);
  if (hot_y) hot_y_int = std::atoi (hot_x);

  cursor = gdk_cursor_new_from_pixbuf  (gdk_display_get_default(),
                                              pixbuf,
                                              hot_x_int,
                                              hot_y_int);
  return cursor;
}

std::string
bmp_skin_get_color (const std::string&	path,
		    const std::string&  key)
{
    std::string filename, value;

    if (Util::find_file (path, BMP_FILE_PLEDIT, filename, 1))
      {
	value = read_ini_string (filename, PLEDIT_TEXT_SECTION_NAME, key);
      }

    return value;
}

GdkColor*
bmp_skin_convert_color (const std::string& value_arg)
{
    GdkColor *color = NULL;

    if (!value_arg.empty ())
      {
	  char *value = g_strdup (value_arg.c_str());
	  char *ptr   = value;
	  int	        len;

	  color = new GdkColor;
	  g_strstrip(value);

	  if (value[0] == '#') ptr++;
	  len = std::strlen(ptr);

	  /*
	   * The handling of incomplete values is done this way
           * to maximize winamp compatibility
	   */
	  if (len >= 6) {
                color->red = hex_chars_to_int(*ptr, *(ptr + 1));
                ptr += 2;
	  }
	  if (len >= 4) {
                color->green = hex_chars_to_int(*ptr, *(ptr + 1));
                ptr += 2;
	  }
	  if (len >= 2) {
                color->blue = hex_chars_to_int(*ptr, *(ptr + 1));
	  }

	  g_free (value);
    }

   return color;
}

void
bmp_colors_set_color (BmpSkinColor color, const std::string& color_hex)
{
    bmp_skin_colors[color] = g_strdup (color_hex.c_str());
}

void
bmp_colors_init (const std::string& archive_dirname)
{
  static const char *names[] = {
   "NormalBG",
   "Normal",
   "SelectedBG",
   "Current"
  };

  for (unsigned int i = 0; i < G_N_ELEMENTS(names); i++)
    {
      bmp_skin_color_names.push_back (names[i]);
    }

  ::bmp_colors_set_color (BmpSkinColor(BMP_SKINCOLOR_PL_BG_NORMAL),   ::bmp_skin_get_color (archive_dirname, ::bmp_skin_color_names[BMP_SKINCOLOR_PL_BG_NORMAL]).c_str());
  ::bmp_colors_set_color (BmpSkinColor(BMP_SKINCOLOR_PL_FG_CURRENT),  ::bmp_skin_get_color (archive_dirname, ::bmp_skin_color_names[BMP_SKINCOLOR_PL_FG_CURRENT]).c_str());
  ::bmp_colors_set_color (BmpSkinColor(BMP_SKINCOLOR_PL_FG_NORMAL),   ::bmp_skin_get_color (archive_dirname, ::bmp_skin_color_names[BMP_SKINCOLOR_PL_FG_NORMAL]).c_str());
  ::bmp_colors_set_color (BmpSkinColor(BMP_SKINCOLOR_PL_BG_SELECTED), ::bmp_skin_get_color (archive_dirname, ::bmp_skin_color_names[BMP_SKINCOLOR_PL_BG_SELECTED]).c_str());
}

void
bmp_colors_free ()
{
    for (unsigned int i = 0; i < SKIN_N_COLORS; i++)
    {
        /* gdk_color_free(bmp_skin_colors_gdk[i]); */
        bmp_skin_colors_gdk[i] = NULL;
    }
}

void
bmp_color_allocate(BmpSkinColor color, GtkWidget *widget)
{
    gdk_color_alloc (gdk_window_get_colormap(widget->window), bmp_colors_get_color(color));
}

GdkColor*
bmp_colors_get_color(BmpSkinColor color)
{
    if (!bmp_skin_colors_gdk[color])
    {
	bmp_skin_colors_gdk[color] = bmp_skin_convert_color(bmp_skin_colors[color]);
    }

    return bmp_skin_colors_gdk[color];
}

void
gdk_color_to_rgba (GdkColor *color, double *r, double *g, double *b, double *a)
{
  *r = color->red   / 65535.0;
  *g = color->green / 65535.0;
  *b = color->blue  / 65535.0;
  *a = 1;
}
